/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1994-1999 Yutaka Sato
Copyright (c) 1994-1999 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, and distribute this material for any purpose
and without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	http.c (HTTP/1.0 proxy)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	940316	created
	99Aug	restructured
//////////////////////////////////////////////////////////////////////#*/
#include "delegate.h"
#include "filter.h"
#include "fpoll.h"
#include "http.h"
#include "vsignal.h"

int KeepAliveWithProxyClient = 0;
int HTTP11_toserver = 1;
int HTTP11_toclient = 1;
int HTTP11_clientTEST = 0;
#define CHUNKED_VER "6.0.0" /* DeleGate version when "chunked" is supported */

int HTTP_ignoreIf = 0;
int HTTP_warnApplet = 0;
int HTTP_rejectBadHeader = 0;

#define clntClose HTTP_clntClose
extern double Time();
extern char *malloc();
extern char *Sprintf();
extern char *CCXcharset();
extern FILE *NULLFP();
extern FILE *TMPFILE();
extern FILE *getTmpFile();
extern FILE *URLget();
extern FILE *hmget1();
extern FILE *cache_fopen_rd();
extern FILE *cache_fopen_rw();
extern FILE *recvDistribution();
extern FILE *openFilter();
extern char *URL_toMyself();
extern char *DDI_fgetsFromC();
extern char *fgetsByBlock();
extern char *file_hostpath();
extern char *paramscanX();
extern char *findFieldValue();
extern char *getFieldValue2();
#define getFieldValue(str,fld,buf,siz) getFieldValue2(str,fld,buf,siz)
#define getFV(str,fld,buf)             getFieldValue2(str,fld,buf,sizeof(buf))

extern int MAX_BUFF_SOCKRECV;
extern int MAX_BUFF_SOCKSEND;
extern int RES_CACHE_DISABLE;
extern int IO_TIMEOUT;
extern int CACHE_TAKEOVER;
extern int URICONV_nFULL;
extern int URICONV_nPARTIAL;
extern int DELEGATE_LastModified;

/*
 *	Connection: Keep-Alive
 */
int HTTP_CKA_PERCLIENT      = 4;
int HTTP_CKA_TIMEOUT        = 5;
int HTTP_CKA_TIMEOUT_MARGIN = 2;
int HTTP_CKA_MAXREQ         = 10;

double HTTP_WAIT_BADSERV = 2000;
double HTTP_WAIT_REQBODY = 30;
double HTTP_POLL_REQBODY = 15;
int REQ_TIMEOUT	= 120;
int HTTP_MAXHOPS = 20;

#define protoGopher(proto)	strcaseeq(proto,"gopher")

#define R_EMPTY_RESPONSE	-10001
#define R_UNSATISFIED		-10002
#define R_UNSATISFIED_TEXT	-10003
#define R_NONTEXT_INTEXT	-10004
#define R_GENERATED		-10005
#define R_MOVED_LOOP		-10006
#define R_BROKEN_RESPONSE	-10007

typedef struct {
	int	 on;
	int	 pid;
	char	*cpath;
	FILE	*cachefp;
	FILE	*tc;
     Connection *Conn;
} relaying;

static struct {
	int	 g_dontTruncate; /* keep the cache in the current state */
	relaying g_relaying;
	char	*g_where;
	int	 g_nsigPipe;
	int	 g_exiting;
} HTTP_current;

#define Relaying	HTTP_current.g_relaying
#define DontTruncate	HTTP_current.g_dontTruncate
#define Where		HTTP_current.g_where
#define nsigPIPE	HTTP_current.g_nsigPipe
#define exiting		HTTP_current.g_exiting

static abortCache()
{	FILE *cachefp;

	cachefp = Relaying.cachefp;
	if( cachefp == NULL )
		return;

	if( Relaying.Conn && Relaying.cpath )
		stopDistribution(Relaying.Conn,cachefp,Relaying.cpath);

	/* discards the half done cache file */
	if( !DontTruncate ){
		sv1log("CACHE discards half written cache.\n");
		fflush(cachefp);
		Ftruncate(cachefp,0,0);
		fseek(cachefp,0,0);
	}
}
static void sigTERM(sig)
{	int now;

	signal(sig,sigTERM);
	now = time(0);
	if( exiting ){
		sv1log("#### duplicate sigTERM in %d seconds\n",now-exiting);
		if( exiting && now-exiting < 10 )
			return;
		else	_Finish(0);
	}
	exiting = now;

	if( Relaying.on ){
		sv1log("HTTP SC got sigTERM(%d) %x\n",sig,Relaying.cpath);
		abortCache();
		if( Relaying.tc != NULL )
			fcloseLinger(Relaying.tc);
		Finish(0); /* flush outputs and exit */
	}else{
		sv1log("HTTP CS got sigTERM(%d)\n",sig);
		_Finish(0);
	}
}
static setRelaying(Conn,tc,cachefp,cpath)
	Connection *Conn;
	FILE *tc,*cachefp;
	char *cpath;
{
	if( Conn == NULL ){
		Relaying.on = 0;
	}else{
		Relaying.on = 1;
		Relaying.pid = getpid();
		Relaying.tc = tc;
		Relaying.cachefp = cachefp;
		Relaying.cpath = cpath;
		Relaying.Conn = Conn;
		Vsignal(SIGTERM,sigTERM);
	}
}
static abortHTTP(msg)
	char *msg;
{
	sv1log("#### EMERGENCY EXIT (%s)\n",msg);
	abortCache();
	_Finish(-1);
}
static void sigPIPE(sig)
{
	signal(SIGPIPE,SIG_IGN); /* sv1log may cause SIGPIPE ... */
	if( nsigPIPE++ % 10 == 0 )
		sv1log("## got SIGPIPE [%d] in HTTP: %s\n",
			nsigPIPE,Where?Where:"");
	signal(SIGPIPE,sigPIPE);
}
static setClientEOF(Conn,tc,fmt,a1,a2,a3)
	Connection *Conn;
	FILE *tc;
	char *fmt,*a1,*a2,*a3;
{	char where[256];

	ClientEOF = 1;
	if( fmt ){
		sprintf(where,fmt,a1,a2,a3);
		sv1log("ClientEOF: %s\n",where);
	}
}
static checkClientEOF(Conn,tc,where)
	Connection *Conn;
	FILE *tc;
	char *where;
{	char *conn_stat;

	if( ClientEOF )
		return ClientEOF;
	if( Conn->from_myself )
		return 0;

	if( ClientSock < 0 ){
		/* maybe this is a internal ResponseFilter process on Win32
		 * spawned without inherited Client-Socket ...
		 */
		return 0;
	}

	if( IsConnected(ClientSock,&conn_stat) )
		return 0;

	if( where ){
		clntClose(Conn,"p:premature client EOF (%s)",where);
		sv1log("## premature client close: %s (%s)\n",where,conn_stat);
	}

	detachFile(tc);
	setClientEOF(Conn,tc,where);

	return 1;
}
static fclosesTIMEOUT(ff,ft)
	FILE *ff,*ft;
{	int df,dt;

	df = fileno(ff);
	dt = fileno(ft);

	fcloseTIMEOUT(ff);
	if( df != dt )
		fcloseTIMEOUT(ft);
	else	fclose(ft);
}

static UAspecific(Conn,UA,direct)
	Connection *Conn;
	char *UA;
{	char *dp;

	if( dp = strstr(UA,"MSIE ") )
		appletFilter = 1;

	if( direct ){
		if( dp = strstr(UA,"MSIE ") ){
			if( dp[5] < '3'
			 || strncmp(dp+5,"3.0",3) == 0 && atoi(dp+8) < 1 )
				clntClose(Conn,"o:old MSIE");
		}else
		if( dp = strstr(UA,"Mozilla/") ){
			if( dp[8] < '2' )
				FlushHead = 50;
			else
			if( dp[8] < '3' && WillKeepAlive )
				FlushIfSmall = 1;
		}
	}

	if( strstr(UA,"Mozilla/0.9") )
		PadCRLF = 1;

	if( dp = strstr(UA,"MSIE ") )
	if( dp[5] < '3' || strncmp(dp+5,"3.0",3)==0 && atoi(dp+8) < 1 )
		SimpleContType = 1;

	if( dp = strstr(UA,"MSIE ") )
	if( REQ_URL[0] != '/' ) /* acting as an origin HTTP server */
	if( URL_toMyself(Conn,REQ_URL) == NULL )
		ProxyAuth = 1;

	if( dp = strstr(UA,"Mosaic") )
	if( dp = strstr(dp,"/") )
	if( strcmp(dp+1,"2.6") < 0 )
		SimpleContType = 1;

	if( dp = strstr(UA,"Lynx") )
	if( dp = strstr(dp,"/") )
		SimpleContType = 1;
}

/* The following stuff is for Mozillas older than 3.0.
 * Either flushHead or Keep-Alive:close is necessary to let their Keep-Alive
 * work properly.
 * - Keep-Alive:close is preferable because their performance by Keep-Alive
 *   with flushHead isn't so good.
 * - But flushHead is preferable when RTT is long...
 */
modifyConnection(Conn,rlength)
	Connection *Conn;
{
	if( FlushIfSmall )
	if( rlength < 1024 )
		HTTP_clntClose(Conn,"o:old Mozilla");

	if( WillKeepAlive )
	if( Conn->xf_filters & (XF_FCL|XF_FTOCL) )
		clntClose(Conn,"x:external filter");
}
getFileUID(fileId,path,peerver)
	char *fileId,*path,*peerver;
{	char host[256],fileid[URLSZ];
	char *addr,*gethostaddr();

	/* if the file is on NFS, should use the name of NFS server */
	GetHostname(host,sizeof(host));

	/* if( peerver < 5.3.1 ) */
		sprintf(fileId,"%s:%s",gethostaddr(host),path);
	/*
	else{
		strcpy(fileid,"file://");
		strcat(fileid,host);
		strcat(fileid,"/");
		strcat(fileid,path);
		toMD5(fileid,fileId);
		sv1log("%s = %s\n",fileId,fileid);
	}
	*/
}
static withConversion(Conn,checkxf)
	Connection *Conn;
{
	/* URL CONVERSIONS */
	if( DO_DELEGATE || IsMounted || URICONV_nFULL || URICONV_nPARTIAL )
		return 1;

	/* CHARACTER CODE CONVERSION */
	if( CCXactive(CCX_TOCL) || cur_codeconvCL(NULL) )
		return 1;

	/* HTML CONVERSIONS */

	/* EXTERNAL FILTERS */
	if( checkxf && Conn->xf_filters )
		return 1;

	return 0;
}
/*
 * Check if some setup of proxy for some conversion have changed
 * after the client cached the data.  If so, whole data should be sent
 * as the response regardless of the modification of original data.
 * So ignore If-Modified-Since to get the body if cache is not.
 * and ignore 304 Not Modified response when the cache is.
 */
static forceBodyResponse(Conn)
	Connection *Conn;
{
	if( !withConversion(Conn,1) )
		return 0;

	if( 0 < ClntIfModClock )
	if( ClntIfModClock < DELEGATE_LastModified ){
		sv1log("DeleGated is Modified after the data: %d > %d\n",
			DELEGATE_LastModified,ClntIfModClock);
		return 1;
	}
	return 0;
}
static LockedByC(Conn,cpath,fc)
	Connection *Conn;
	char *cpath;
	FILE *fc;
{	char fileId[URLSZ];

	if( cpath == NULL || cpath[0] == 0 )
		return -1;
	if( LockedByClient == 0 )
		return 0;

	getFileUID(fileId,cpath);
	if( streq(LockedByClient,fileId) ){
		sv1log("--- cache is locked by client: %s\n",cpath);
		DontUseCache = 1;
		DontReadCache = DontWriteCache = DontWaitCache = 1;
		return 1;
	}
	return 0;
}
/*
 * generate `If-Modified-Since: cdate' if not specified by the client.
 */
static genIfModified(Conn,cachefp,cpath,cdate)
	Connection *Conn;
	FILE *cachefp;
	char *cpath;
{	char scdate[64],mod[64];
	int lastmod;

	if(forceBodyResponse(Conn)
	 && cachefp == NULL
	 && !LockedByC(Conn,cpath,NULL)){
		/* need data anyway to produece the body response regardless
		 *  of the modified date, so ignore and don't produce If-Mod.
		 */
		sv1log("ignore and don't produce If-Modified-Since\n");
		return;
	}
	if(cpath==NULL||cdate==-1||cachefp==NULL||DontUseCache||DontReadCache){
		/* local cache is not useful, pass Client's request as is */
		if( ClntIfMod[0] )
			RFC822_addHeaderField(REQ_FIELDS,ClntIfMod);
		return;
	}

	/*
	 * local cache file is available
	 */
	lastmod = HTTP_getLastModInCache(scdate,sizeof(scdate),cachefp,cpath);

	if( ClntIfMod[0] && lastmod <= ClntIfModClock ){
		RFC822_addHeaderField(REQ_FIELDS,ClntIfMod);
		CacheLastMod = ClntIfModClock;
		return;
	}

	if( lastmod == 0 )
		lastmod = HTTP_genLastMod(scdate,sizeof(scdate),cachefp,cpath);
	CacheLastMod = lastmod;

	sprintf(mod,"If-Modified-Since: %s\r\n",scdate);
	RFC822_addHeaderField(REQ_FIELDS,mod);
	Verbose("+ %s",mod);
}
static setReferer(Conn,proto,host,port,req,referer,refbuf)
	Connection *Conn;
	char *proto,*host,*req;
	Referer *referer;
	char *refbuf;
{	char *rp,*bp,*up;

	rp = refbuf;

	referer->r_my.u_hostport = rp;
	HTTP_ClientIF_HP(Conn,rp);
	rp += strlen(rp) + 1;

	referer->r_my.u_proto = CLNT_PROTO;
	referer->r_my.u_host = rp;
	referer->r_my.u_port = HTTP_ClientIF_H(Conn,rp);
	rp += strlen(rp) + 1;

	if( REMOTE_HOST[0] ){
		proto = REMOTE_PROTO;
		host = REMOTE_HOST;
		port = REMOTE_PORT;
	}

	referer->r_sv.u_proto = rp;
	strcpy(rp,proto);
	rp += strlen(rp) + 1;

	referer->r_sv.u_host = rp;
	strcpy(rp,host);
	rp += strlen(rp) + 1;

	referer->r_sv.u_port = port;

	referer->r_sv.u_hostport = rp;
	HostPort(rp,proto,host,port);
	rp += strlen(rp) + 1;

	referer->r_sv.u_method = rp;
	up = wordscanX(req,rp,32);
	rp += strlen(rp) + 1;
	
	referer->r_sv.u_path = rp;
	wordscanX(up,rp,1024);
	/*if( toProxy )*/
	strip_urlhead(rp,NULL,NULL);
	rp += strlen(rp) + 1;

	referer->r_sv.u_base = rp;
	strcpy(rp,referer->r_sv.u_path);
	if( bp = strrchr(rp,'/') )
		bp[1] = 0;
	else	rp[0] = 0;
	rp += strlen(rp) + 1;

	referer->r_my.u_path = rp;
	strcpy(rp,Conn->cl_baseurl);
	rp += strlen(rp) + 1;

	referer->r_vb = Conn->my_vbase; /* when BASEURL is a FULL-URL */

	referer->r_altbuf = rp;
}
HTTP_delayReject(Conn,req,stat,self)
	Connection *Conn;
	char *req,*stat;
{	char *method,*url;
	HttpRequest reqx;
	char path[URLSZ],referer[URLSZ];
	char shost[256];
	int sport;

	decomp_http_request(req,&reqx);
	method = reqx.hq_method;
	url = reqx.hq_url;

	path[0] = 0;
	decomp_absurl(url,NULL,NULL,path,sizeof(path));
	HTTP_getRequestField(Conn,"Referer",referer,sizeof(referer));

	self |= DO_DELEGATE || IsMounted || IAM_GATEWAY;
	sport = getClientHostPort(Conn,shost);
	logReject(Conn,self,shost,sport);
	delayReject(Conn,self,method,DFLT_PROTO,shost,sport,path,referer,stat);
}
HTTP_delayCantConn(Conn,req,stat,self)
	Connection *Conn;
	char *req,*stat;
{	char method[64];
	char shost[256];
	int sport;

	wordScan(req,method);

	self |= DO_DELEGATE || IsMounted || IAM_GATEWAY;
	sport = getClientHostPort(Conn,shost);
	delayReject(Conn,self,method,DFLT_PROTO,shost,sport,"","",stat);
}

/*###################################### RESPONSE */
typedef struct {
	FILE   *t_fp;		/* FILE pointer */
	int	t_fd;		/* file descriptor */
	int	t_EOF;		/* (soft) EOF reached */
	int	t_buffered;	/* bytes in buffer */
	int	t_nready;	/* I/O ready */
	int	t_lastRcc;	/* bytes read in the last */

	int	t_contLeng;	/* Content-Length got / put */
	int	t_headTotal;	/* transferred header size */
	int	t_bodyTotal;	/* transferred body size */
	int	t_keepAlive;	/* Connection: keep-alive or HTTP/1.1 */

	int	t_remsize;	/* remaining (input) body size */
	char	t_buffer[IBUFSIZE+1];

	int	t_chunked;	/* is chunked */
	int	t_chunk_ser;	/* serial number of the current chunk */
	int	t_chunk_siz;	/* size of the current chunk */
	int	t_chunk_rem;	/* remaining bytes of the current chunk */
} TransCode;

typedef struct {
	Connection *r_Conn;
	Referer *r_referer;

	int	r_qWithHeader;
	char	r_qmethod[128];	/* method of the request */
	int	r_reqHEAD;	/* qmethod is HEAD */

   HttpResponse	r_status;		/* decomposed status line */
	char	r_connection[256];	/* Connection: */
	char	r_ctype[256];		/* Content-Type: */
	char	r_servername[256];	/* Server: */
	char	r_cencoding[256];	/* Content-Encoding: */
	char	r_tencoding[256];	/* Transfer-Encoding: */
	char	r_setCookie[2048];	/* Set-Cookie: */
	char	r_cachecontrol[256];	/* Cache-Control: */

	int	r_lastMod;	/* with Last-Modified: */
	int	r_lastModTime;	/* Last-Modified: */

	int	r_errori;
	int	r_tryKeepAlive;

	int	r_convChar;	/* with character conversion */
	int	r_deEnt;	/* decode char. entitiy encoding &xx; */
	int	r_putPRE;	/* with <PRE> body </PRE> insertion */

	int	r_noBody;	/* response without body */
	int	r_inHeader;	/* processing header part of resp. */
	int	r_isText;	/* type of text */
	int	r_isBin;	/* body is binary data */

	int	r_bodyLines;	/* line# of text body got from server */
	int	r_txtLen;
	int	r_binLen;

	double	r_Start;	/* begining of response relay */
	double	r_Resp1;	/* first response */
	double	r_firstResp;
	double	r_connDelay;

	int	r_nflush;	/* flush count */
	int	r_ninput;
	int	r_noutput;
	int	r_nsigreport;

	TransCode r_fsx;	/* transfer coding from server */
	TransCode r_tcx;	/* to client */

	char   *r_tmpFileId;
	FILE   *r_respBuff;	/* tmporary file to receive response */
	FILE   *r_tc_sav;	/* saved original FILE to client */

	int	r_niced;
	FILE   *r_cachefp;
	char   *r_cpath;
	struct {
	  int	d_start;
	  char	d_cpath[URLSZ];
	  FILE *d_fp;
	  int	d_ready;
	} r_dcx;
} ResponseContext;

#define RX_Conn		RX->r_Conn
#define RX_referer	RX->r_referer
#define RX_qWithHeader	RX->r_qWithHeader
#define RX_qmethod	RX->r_qmethod
#define RX_reqHEAD	RX->r_reqHEAD

#define RX_status	RX->r_status
#define RX_isHTTP09	RX->r_status.hr_isHTTP09
#define RX_ver		RX->r_status.hr_ver
#define RX_code		RX->r_status.hr_rcode
#define RX_reason	RX->r_status.hr_reason

#define RX_connection	RX->r_connection
#define RX_ctype	RX->r_ctype
#define RX_servername	RX->r_servername
#define RX_cencoding	RX->r_cencoding
#define RX_tencoding	RX->r_tencoding
#define RX_setCookie	RX->r_setCookie
#define RX_cachecontrol	RX->r_cachecontrol

#define RX_lastMod	RX->r_lastMod
#define RX_lastModTime	RX->r_lastModTime

#define RX_errori	RX->r_errori
#define RX_tryKeepAlive	RX->r_tryKeepAlive

#define RX_convChar	RX->r_convChar
#define RX_deEnt	RX->r_deEnt
#define RX_putPRE	RX->r_putPRE

#define RX_noBody	RX->r_noBody
#define RX_inHeader	RX->r_inHeader
#define RX_isText	RX->r_isText
#define RX_isBin	RX->r_isBin

#define RX_bodyLines	RX->r_bodyLines
#define RX_txtLen	RX->r_txtLen
#define RX_binLen	RX->r_binLen

#define RX_Start	RX->r_Start
#define RX_Resp1	RX->r_Resp1
#define RX_firstResp	RX->r_firstResp
#define RX_connDelay	RX->r_connDelay

#define RX_nflush	RX->r_nflush
#define RX_ninput	RX->r_ninput
#define RX_noutput	RX->r_noutput
#define RX_nsigreport	RX->r_nsigreport

#define RX_fsx		RX->r_fsx
#define RX_fsp		RX->r_fsx.t_fp
#define RX_fsd		RX->r_fsx.t_fd
#define RX_rdBuff	RX->r_fsx.t_buffer
#define RX_rdContLen	RX->r_fsx.t_contLeng
#define RX_rdHeadTotal	RX->r_fsx.t_headTotal
#define RX_rdTotal	RX->r_fsx.t_bodyTotal
#define RX_remsize	RX->r_fsx.t_remsize

#define RX_tc_sav	RX->r_tc_sav
#define RX_tmpFileId	RX->r_tmpFileId
#define RX_respBuff	RX->r_respBuff

#define RX_tcx		RX->r_tcx
#define RX_tcp		RX->r_tcx.t_fp
#define RX_tcd		RX->r_tcx.t_fd
#define RX_wrbufed	RX->r_tcx.t_buffered
#define RX_wrHeadTotal	RX->r_tcx.t_headTotal
#define RX_wrContLen	RX->r_tcx.t_contLeng
#define RX_wrTotal	RX->r_tcx.t_bodyTotal
#define RX_niced	RX->r_niced

#define RX_cachefp	RX->r_cachefp
#define RX_cpath	RX->r_cpath
#define RX_dcx		RX->r_dcx

static url_abs_delegateS(Conn,do_delegate,referer,src,dst)
	Connection *Conn;
	Referer *referer;
	char *src,*dst;
{	char buf[0x10000];

	/*if( do_delegate || something-MOUNTed )*/
		url_absoluteS(referer,src,buf);
	CTX_url_delegateS(Conn,referer,buf,dst,do_delegate?Conn:NULL);

	if( DO_DELEGATE || IsMounted )
	if( URICONV_nPARTIAL )
	if( RespCode != 301 && RespCode != 302 )
	if( url_partializeS(referer,dst,buf) )
		strcpy(dst,buf);
}
static redirect_HTMLS(Conn,dont_rewrite,do_delegate,referer,src,dst)
	Connection *Conn;
	Referer *referer;
	char *src,*dst;
{
	if( dont_rewrite )
		url_absoluteS(referer,src,dst);
	else	url_abs_delegateS(Conn,do_delegate,referer,src,dst);
}
redirect_HTML0(Conn,proto,host,port,req,src,dst)
	Connection *Conn;
	char *proto,*host,*req;
	char *src,*dst;
{	Referer referer;
	char refbuf[URLSZ];

	setReferer(Conn,proto,host,port,req,&referer,refbuf);
	redirect_HTMLS(Conn,DONT_REWRITE,DO_DELEGATE,&referer,src,dst);
}

#define Line(line0,ln)	((ln)==0?line0:((ln)%2==0?Line0:Line1))
#define Cline	Line(line0,linex+0)
#define Nline	Line(line0,linex+1)
static char *metahttp(Conn,attr,cont,tagp,contp,nextp)
	Connection *Conn;
	char *attr,*cont,*tagp,*contp,*nextp;
{	char *cs,charset[64];
	int len;

	if( strcaseeq(attr,"Content-Type") )
	if( len = erase_charset_param(cont,NULL) ){
		len = erase_charset_param(contp,charset);
		sv1log("removed charset in HTTP-EQUIV[%s][%s]\n",cont,charset);
		return nextp - len;
	}
	return nextp;
}
static char *rewriteResponse(Conn,referer,line0,ctype,doconv,deent,tc)
	Connection *Conn;
	Referer *referer;
	char *line0,*ctype;
	FILE *tc;
{	int linex;

	linex = 0;

	if( !RelayTHRU )
	if( BORN_SPECIALIST || ACT_SPECIALIST ){
		if( ctype == NULL )
			ctype = "message/rfc822";

		if( anchor_rem[0] ){
			strcpy(Nline,anchor_rem);
			strcat(Nline,Cline);
			linex++;
			anchor_rem[0] = 0;
		}
		if( !DONT_REWRITE || URICONV_nFULL )
		if( html_nextTagAttr(Cline,ctype,anchor_rem,NULL,NULL,NULL) ){
			redirect_HTMLS(Conn,DONT_REWRITE,DO_DELEGATE,referer,Cline,Nline);
			linex++;
		}
		if( deent ){
			decode_entities(Cline,Nline);
			linex++;
		}
		if( doconv ){
			scan_metahttp(Conn,Cline,metahttp);
			if( CCXactive(CCX_TOCL) )
				CCXexec(CCX_TOCL,Cline,strlen(Cline),
					Nline,RESP_LINEBUFSZ);
			else	line_codeconv(Cline,Nline,ctype);
			linex++;
		}
	}
	return Cline;
}

#define TX_PLAIN	1
#define TX_HTML		2
#define TX_MISC		3

#define STATIC
STATIC
char *GETRESPTXT(Conn,RX,fromcache)
	Connection *Conn;
	ResponseContext *RX;
{	int timeout;
	int remleng;
	int niced = RX_niced;
	FILE *fs = RX_fsp;
	char *line = RX_rdBuff;
	int byline = RX_inHeader;
	int size = RX_remsize + 1;
	int *lengp = &RX_fsx.t_lastRcc;
	int *isbinp = &RX_isBin;
	int cc;
	char *rcode;

	timeout = IO_TIMEOUT*1000;
	if( !RX_inHeader && 0 < RX_rdContLen ){
		remleng = RX_rdContLen - RX_rdTotal;
		if( !fromcache ){
			if( remleng <= 0 )
				timeout = 3*1000;
		}
	}else	remleng = RX_remsize;

	rcode = fgetsByBlock(line,size,fs,niced,timeout,byline,fromcache,
		remleng,lengp,isbinp);
	cc = *lengp;
	if( RESP_SAV ){
		bcopy(line,RESP_MSG+RESP_LEN,cc+1);
		RESP_LEN += cc;
	}
	return rcode;
}

/*
fflush after the header is put makes separated header/body and let
the server on Solaris2.X be slower...  But, this is necessary in
BSDI-1.x to avoid the header from being put after the body X-<
*/
#ifdef __bsdi__
#include <sys/param.h>
#if !defined(_BSDI_VERSION) || _BSDI_VERSION < 199501
#define HeadFlush 1
#endif
#endif
#ifndef HeadFlush
#define HeadFlush 0
#endif

STATIC flushHead(Conn,tc,length)
	Connection *Conn;
	FILE *tc;
{
/*
	if( 0 < length && length < 1024 )
*/
 	if( !HeadFlush && !FlushHead )
		return 0;

	if( fflush(tc) == EOF ){
		setClientEOF(Conn,tc,"flushHead");
		return EOF;
	}

	Verbose("HeadFlush:%d, FlushHead:%d(msec)\n",HeadFlush,FlushHead);
	if( FlushHead )
		msleep(FlushHead);

	return 0;
}
HTTP_putMIMEmsg(Conn,in,out)
	Connection *Conn;
	FILE *in,*out;
{	char line[1024];
	int hsize,msize;
	int chunked = 0;
	int oleng,nleng,wleng;
	/* SHOULD rewrite "Line:" header also */
	int fnlen;

	oleng = 0;
	for(;;){
		if( fgets(line,sizeof(line),in) == NULL ) break;
		if( line[0] == '\r' || line[0] == '\n' ) break;
		if( fnlen = STRH(line,F_ContLeng) )
			oleng = atoi(line+fnlen);
		else
		if( fnlen = STRH(line,F_TransEncode) ){
			if( strcasestr(line+fnlen,"chunked") )
				chunked = 1;
			sv1log("#HT11 chunked=%d %s\n",chunked,line);
		}
	}
	msize = file_size(fileno(in));
	hsize = ftell(in);
	nleng = msize - hsize;
	sv1log("Content-Length: %d -> %d (%d - %d)\n",oleng,nleng,msize,hsize);

	fseek(in,0,0);
	for(;;){
		if( fgets(line,sizeof(line),in) == NULL ) break;
		if( line[0] == '\r' || line[0] == '\n' ){
			if( !chunked )
			fprintf(out,"%s %d\r\n",F_ContLeng,nleng);
			fputs(line,out);
			break;
		}
		if( fnlen = STRH(line,F_ContLeng) ) continue;
		fputs(line,out);
	}

	if( Conn && flushHead(Conn,out,nleng) != 0 ){
 sv1log("## stop relay-3: flush()=EOF, client seems dead\n");
		return -1;
	}else{
		wleng = copy_fileTIMEOUT(in,out,NULL);
		if( wleng != nleng ){
 sv1log("## stop relay-4: FWRITE(%d)=%d, SIGPIPE=%d, client seems dead\n",
 nleng,wleng,nsigPIPE);
			return -1;
		}
	}
	return wleng;
}
static FILE *detach_respbuff(Conn,tc_buf,tc,checkHead,reason)
	Connection *Conn;
	FILE *tc_buf,*tc;
	char *reason;
{	char field[1024],fname[1024],fbody[1024];

	sv1log("detach respBuff: %s\n",reason);

	fflush(tc_buf);
	Ftruncate(tc_buf,0,1);
	fseek(tc_buf,0,0);

	if( checkClientEOF(Conn,tc,"detach_respbuff") )
		httpStat = CS_EOF;

	/* copy header removing Connection: Keep-Alive */
	if( checkHead )
	while( fgets(field,sizeof(field),tc_buf) ){
		if( STRH_Connection(field) ){
			scan_field1(field,fname,sizeof(fname),fbody,sizeof(fbody));
			if( strncasecmp(fbody,"keep-alive",10) == 0 )
				clntClose(Conn,"b:temp. buff. detached (%s)",
					reason);
			sprintf(field,"%s: close\r\n",fname);
			SentKeepAlive = 0;
		}
		if( !ClientEOF )
		if( fputs(field,tc) == EOF )
			setClientEOF(Conn,tc,"detach_respbuff-1");

		if( field[0] == '\r' || field[0] == '\n' )
			break;
	}

	if( !ClientEOF ){
		copy_fileTIMEOUT(tc_buf,tc,NULL);
		if( feof(tc) )
			setClientEOF(Conn,tc,"detach_respbuff-2");
	}
	fclose(tc_buf);
	return tc;
}
static lastmtime(line)
	char *line;
{	char scdate[1024];

	lineScan(line,scdate);
	return scanHTTPtime(scdate);
}
static scanRespHead1(Conn,RX,line)
	Connection *Conn;
	ResponseContext *RX;
	char *line;
{	int fnlen;

	if( fnlen = STRH(line,F_ContEncode) )
		wordScan(line+fnlen,RX_cencoding);
	else
	if( fnlen = STRH(line,F_ContLeng) )
		RX_rdContLen = atoi(line+fnlen);
	else
	if( fnlen = STRH(line,F_ContRange) ){
		sv1log("#HT11 %s",line);
	}else
	if( fnlen = STRH(line,F_LastMod) ){
		RX_lastMod = 1;
		RX_lastModTime = lastmtime(line+fnlen);
	}else
	if( fnlen = STRH(line,F_ContType) ){
		char *pp;
		wordScan(line+fnlen,RX_ctype);
		if( pp = strchr(RX_ctype,';') )
			*pp = 0;
		if( STRH(line,F_CtypeText) ){
			if( STRH(line,F_CtypePlain) ){
				if( plain2html() ){
					RX_isText = TX_HTML;
					RX_putPRE = 1;
				}else	RX_isText = TX_PLAIN;
			}else
			if( STRH(line,F_CtypeHTML) )
				RX_isText = TX_HTML;
			else	RX_isText = TX_MISC;
		}
	}else
	if( fnlen = STRH(line,F_Server) ){
		lineScan(line+fnlen,RX_servername);
if(LOG_GENERIC){
fputLog(Conn,"Server","%s://%s:%d; version=%s\n",
DST_PROTO,DST_HOST,DST_PORT,RX_servername);
}
	}else
	if( fnlen = STRH(line,F_CacheControl) ){
		lineScan(line+fnlen,RX_cachecontrol);
	}else
	if( line[0] == '\r' || line[0] == '\n' ){
		RX_inHeader = 0;
	}
	if( RX_errori )
	if( RX_code != 404 )
		sv1log("HTTP error header: %s",line);
}

/*
 * do something before ending response header
 * with empty CRLF line
 */
static endRespHead(Conn,RX)
	Connection *Conn;
	ResponseContext *RX;
{	char via[256];

	if( !ImResponseFilter ){
		if( !ClientEOF && RX_qWithHeader ){
			HTTP_genVia(Conn,0,via);
			fprintf(RX_tcp,"Via: %s\r\n",via);
		}
	}

	HTTP_editResponseHeader(Conn,RX_tcp);
	sv1log("#HT11 SERVER ver[%s] conn[%s]\n",RX_ver,RX_connection);

	ServKeepAlive = 0;
	if( !(toMaster && vercmp(MediatorVer,CHUNKED_VER) < 0) )
	if( streq(RX_ver,"HTTP/1.1") && strncasecmp(RX_connection,"close",6) != 0
	 || strncasecmp(RX_connection,"keep-alive",10) == 0
	){
		ServKeepAlive = 1;
		RX_fsx.t_keepAlive = 1;
		sv1log("#HT11 server KEEP-ALIVE\n");
	}

	/* asked Keep-Alive in the request, so must reply
	 * int the respose.
	 */
	if( !ClientEOF && ClntKeepAlive ){
		if( !RX_noBody )
			modifyConnection(Conn,RX_rdContLen);

		if( RX_tryKeepAlive ){
/*
if( !RX_noBody && RX_tcp != RX_respBuff && RX_rdContLen==0 )
clntClose(Conn,"u:unbound length data");
*/
			if( RX_errori )
				clntClose(Conn,"s:bad status: %d",RX_errori);

			if( Conn->xf_filters & (XF_FTOCL|XF_FCL) )
				clntClose(Conn,"x:external filter");
		}
		putKeepAlive(Conn,RX_tcp);
	}

	if( !ClientEOF ){
		if( RX_tcx.t_chunked && !RX_noBody ){
			sv1log("#HT11 --putChunk-Header: %s chunked\r\n",
				F_TransEncode);
			fprintf(RX_tcp,"%s chunked\r\n",F_TransEncode);
		}
	}
}
static putChunk(RX,buff,leng)
	ResponseContext *RX;
	char *buff;
{	int wcc,chunked;

	if( RX_isHTTP09 )
		chunked = 0;
	else	chunked = !RX_inHeader && RX_tcx.t_chunked;
	if( chunked ){
		RX_tcx.t_chunk_ser += 1;
		Verbose("#HT11 --putChunk[%d] %d\n",RX_tcx.t_chunk_ser,leng);
		fprintf(RX_tcp,"%x\r\n",leng);
	}
	if( 0 < leng )
		wcc = fwriteTIMEOUT(buff,1,leng,RX_tcp);
	else	wcc = 0;
	if( chunked ){
		fputs("\r\n",RX_tcp);
	}
	return wcc;
}
static getChunk(RX)
	ResponseContext *RX;
{	char line[1024];

	if( 0 < RX_fsx.t_chunk_ser ){
		while( fgets(line,sizeof(line),RX_fsp) != NULL ){
			Verbose("#HT11 --getChunk[%d] footer: %s",
				RX_fsx.t_chunk_ser,line);
			if( *line == '\r' ) /* CRLF */
				break; 
				/* entitiy-headers in the last chunk */
		}
		if( RX_fsx.t_chunk_siz == 0 )
			return -1;
	}
	line[0] = 0;
	fgets(line,sizeof(line),RX_fsp);
	RX_fsx.t_chunk_rem = 0;
	sscanf(line,"%x",&RX_fsx.t_chunk_siz);
	RX_remsize = RX_fsx.t_chunk_siz;
	if( sizeof(RX_rdBuff)-1 < RX_remsize )
		RX_remsize = sizeof(RX_rdBuff)-1;
	RX_fsx.t_chunk_rem = RX_fsx.t_chunk_siz; 
	RX_fsx.t_chunk_ser++;

	Verbose("#HT11 --getChunk[%d] header: (%d) %s",RX_fsx.t_chunk_ser,
		RX_fsx.t_chunk_siz,line);
	return RX_fsx.t_chunk_ser;
}
static fputsResponseChunk(Conn,RX,line,ctype,doconv,deent)
	Connection *Conn;
	ResponseContext *RX;
	char *line,*ctype;
{	char *oline;
	int leng,wcc;

	if( RelayTHRU || !BORN_SPECIALIST && !ACT_SPECIALIST ){
		oline = line;
		leng = strlen(line);
	}else{
	oline = rewriteResponse(Conn,RX_referer,line,ctype,doconv,deent,RX_tcp);
	leng = strlen(oline);
	}

	if( 0 < leng ){
		wcc = putChunk(RX,oline,leng);
		if( wcc < leng ){
			setClientEOF(Conn,RX_tcp,"fputsResponse");
			return 0;
		}else	return wcc;
	}else{
		sv1log("#HT11 fputsResponse(leng=%d)\n",leng);
		return leng;
	}
}
#define FPUTS(line,ctype,doconv,deent) \
	fputsResponseChunk(Conn,RX,line,RX_inHeader?0:ctype,doconv,deent)

static fwriteResponseChunk(Conn,RX,buff,leng)
	Connection *Conn;
	ResponseContext *RX;
	char *buff;
{	int wcc;

	if( leng <= 0 )
		return leng;
	return putChunk(RX,buff,leng);
}
char *HTTP_outCharset(Conn)
	Connection *Conn;
{	char *xcharset;

	xcharset = 0;
	if( CCXactive(CCX_TOCL) )
		xcharset = CCXcharset(CCX_TOCL);
	else	codeconv_get(NULL,&xcharset,NULL);
	return xcharset;
}

extern int DHTML_printConn();
static filterApplet(Conn,RX,line)
	Connection *Conn;
	ResponseContext *RX;
	char *line;
{	FILE *tmp;
	char *dp,userAgent[64],warn[URLSZ];
	int rcc;

	dp = strcasestr(line,"<APPLET ");
	if( dp == NULL )
		return 0;

	if( strstr(proxyCookie,"APPLET") ){
		sv1log("#DONT CHECK APPLET: Cookie=%s\n",proxyCookie);
		return 0;
	}

	strcpy(userAgent,"UA-TEST");
	tmp = TMPFILE("AppletWarning");
	putBuiltinHTML(Conn,tmp,"AppletWarning","applet.dhtml",NULL,
		DHTML_printConn,NULL);
	fflush(tmp); fseek(tmp,0,0);
	rcc = fread(warn,1,sizeof(warn),tmp);
	if( rcc < 0 )
		rcc = 0;
	warn[rcc] = 0;
	fclose(tmp);
	Strins(dp,warn);
	daemonlog("E","#Inserted Warning for APPLET\n");
}

static relayTxtResp1(Conn,RX,req)
	Connection *Conn;
	ResponseContext *RX;
	char *req;
{	int wcc;
	char *line = RX_rdBuff;
	int fnlen;

	if( !RX_inHeader ){
		if( HTTP_warnApplet )
		if( appletFilter )
			filterApplet(Conn,RX,line);

		wcc = FPUTS(line,RX_ctype,!RX_inHeader && RX_convChar,RX_deEnt);
		RX_wrTotal += wcc;
		return;
	}

	if( SimpleContType && STRH(line,F_ContType) ){
		char *dp;
		if( dp = strchr(line,';') ){
			sv1log("Discard Parameter -- %s",dp);
			strcpy(dp,"\r\n");
		}
	}

	if( RX_deEnt && STRH(line,F_DGVer) )
		RX_deEnt = 0;

	/*
	 * the charset should be rewritten only when
	 * the code conversion in the body is made ...
	 */
	if( RX_convChar && !SimpleContType && STRH(line,F_CtypeText) ){
		char *xcharset;
		if( xcharset = HTTP_outCharset(Conn) )
			replace_charset(line,xcharset);
	}

	if( fnlen = STRH(line,F_ContLeng) ){
		if( RX_tcx.t_chunked ){
			sv1log("#HT11 chunked, should skip: %s",line);
			/*return;*/
		}
	}else
	if( fnlen = STRH(line,F_SetCookie) ){
		lineScan(line+fnlen,RX_setCookie);
		MountCookieResponse(Conn,req,line+fnlen);
		if( RX_cachefp ){
			Ftruncate(RX_cachefp,0,0);
			fseek(RX_cachefp,0,0);
			RX_cachefp = NULL;
		}
	}else
	if( fnlen = STRH(line,F_Location) ){
		if( RX_code == 301 || RX_code == 302 ){
			sv1log("%s",line);
			if( isMovedToSelf(Conn,line+fnlen) ){
				RX_errori = R_MOVED_LOOP;
				sv1log("#### don't cache MOVED loop\n");
			}
			if( WillKeepAlive && isMovedToAnotherServer(Conn,line+fnlen) )
				clntClose(Conn,"m:moved to another server");
		}
	}

	if( HTTP_ignoreIf && (
	    (fnlen = STRH(line,F_Etag))
	 || (fnlen = STRH(line,F_LastMod))
	)){
		sv1log("IGNORE IF-: %s",line);
	}else
	if( fnlen = STRH_Connection(line) ){
		Verbose("HCKA:[R] %s",line);
		lineScan(line+fnlen,RX_connection);
	}else
	if( fnlen = STRH(line,F_TransEncode) ){
		wordScan(line+fnlen,RX_tencoding);
		if( strcaseeq(RX_tencoding,"chunked") ){
			sv1log("#HT11 --getChunk-Header: %s",line);
			RX_fsx.t_chunked = 1;
		}
	}else
	if( RX_qWithHeader && plain2html() && STRH(line,F_CtypePlain) ){
		if( !ClientEOF )
			fprintf(RX_tcp,"%s\r\n",F_CtypeHTML);
	}else
	if( RX_qWithHeader ){
	    if( !ClientEOF ){
		wcc = FPUTS(line,RX_ctype,!RX_inHeader && RX_convChar,RX_deEnt);
		RX_wrHeadTotal += wcc;
	    }
	}
}
static GETRESPBIN(Conn,RX,fromcache,nready)
	Connection *Conn;
	ResponseContext *RX;
{	int rcc;
	char *buff = RX_rdBuff;

	if( fromcache ){
		rcc = fread(buff,1,RX_remsize,RX_fsp);
		return RX_fsx.t_lastRcc = rcc;
	}
	if( rcc = fgetBuffered(buff,RX_remsize,RX_fsp) ){
		Verbose("buffered(%d)\n",rcc);
		return RX_fsx.t_lastRcc = rcc;
	}

	/* MASTER-DeleGate (5.X) may keep connection even after said
	 * "Connection:close" ...
	  if( RX_rdTotal == RX_rdContLen )
	  if( nready == 0 && (nready = PollIn(fileno(RX_fsp),1000)) == 0 ){
		sv1log("#### SOFT EOF %d/%d %s,%s\n",RX_rdTotal,RX_rdContLen,
			RX_ver,RX_connection);
		RX_fsx.t_EOF = 1;
		return RX_fsx.t_lastRcc = 0;
	  }
	 */

	if( nready == 0 && (nready = PollIn(fileno(RX_fsp),2000)) == 0 ){
		if( RX_tcp == RX_respBuff ){
			RX_tcp = detach_respbuff(Conn,RX_tcp,RX_tc_sav,
				RX_qWithHeader,"flush slow binary body.");
		}else
		if( !ClientEOF ){
			RX_nflush++;
			if( fflushTIMEOUT(RX_tcp) == EOF )
				setClientEOF(Conn,RX_tcp,"binary");
		}
	}

	Verbose("#HT11 READ-BIN: %d [%d/%d]\n",
		RX_remsize,RX_rdTotal,RX_rdContLen);

	if( nready == 0 )
		rcc = readTIMEOUT(fileno(RX_fsp),buff,RX_remsize);
	else	rcc = read(fileno(RX_fsp),buff,RX_remsize);

	if( rcc <= 0 )
		RX_fsx.t_EOF = 1;
	else
	while( rcc < RX_remsize ){
	    if( 0 < PollIn(fileno(RX_fsp),10) ){
		int rcc1;
		rcc1 = read(fileno(RX_fsp),buff+rcc,RX_remsize-rcc);
		if( rcc1 <= 0 ){
			RX_fsx.t_EOF = 1;
			break;
		}else{
			rcc += rcc1;
			RX_ninput++;
		}
	    }else	break;
	}
	return RX_fsx.t_lastRcc = rcc;
}

static fflushKeepAlive(Conn,where,fc,tc,wtotal)
	Connection *Conn;
	char *where;
	FILE *fc,*tc;
{	int nready;

	if( fc == NULL ){
		if( fflushTIMEOUT(tc) == EOF )
			setClientEOF(Conn,fc,"%s-fflushKeepAlive-1",
				where);
	}else
	if( wtotal == 0 || 1024*8 < wtotal ){
		nready = fPollIn(fc,1);
		if( nready < 0 )
			setClientEOF(Conn,fc,"%s-fflushKeepAlive-2-RESET",
				where);
		else
		if( nready == 0 ){
			set_nodelay(fileno(tc),1);
			if( fflushTIMEOUT(tc) == EOF )
				setClientEOF(Conn,fc,"%s-fflushKeepAlive-3",
					where);
			else	set_nodelay(fileno(tc),0);
		}
	}
}
static abortKeepAlive(Conn,RX)
	Connection *Conn;
	ResponseContext *RX;
{
	if( ClientEOF ){
	clntClose(Conn,"r:by client (response EOF) %d/%d (%d/%d)",
		RX_wrTotal,RX_wrContLen,RX_rdTotal,RX_rdContLen);
	}else
	if( !RX_tcx.t_chunked && !RX_noBody && RX_wrTotal != RX_wrContLen ){
	clntClose(Conn,"l:##FATAL:inconsistent body length %d/%d (%d/%d)",
		RX_wrTotal,RX_wrContLen,RX_rdTotal,RX_rdContLen);
	}else
	if( RX_errori ){
	clntClose(Conn,"s:##FATAL:bad status(%d) %d/%d (%d/%d)",
		RX_errori,RX_wrTotal,RX_wrContLen,RX_rdTotal,RX_rdContLen);
	}
}

static releaseRespbuf(Conn,RX)
	Connection *Conn;
	ResponseContext *RX;
{
	fflush(RX_tcp);
	Ftruncate(RX_tcp,0,1);
	fseek(RX_tcp,0,0);

	if( checkClientEOF(Conn,RX_tcp,"flush_respbuff") ){
		httpStat = CS_EOF;
		RX_wrTotal = RX_wrContLen = 0;
	}else
	if( 0 < (RX_wrTotal = HTTP_putMIMEmsg(Conn,RX_tcp,RX_tc_sav)) ){
		RX_wrContLen = RX_wrTotal;
	}else{
		sv1log("## HTTP RESPONSE CLOSE: NO (client dead)\n");
		setClientEOF(Conn,RX_tcp,"flush_respbuff.FAILED");
		freeTmpFile(RX_tmpFileId,Conn);
	}
	fclose(RX_tcp);
	RX_tcp = RX_tc_sav;
}
static relayBinResp1(Conn,RX,rcc)
	Connection *Conn;
	ResponseContext *RX;
{	int wcc;
	char *buff = RX_rdBuff;

	if( rcc <= 0 ){
		Verbose("read(%d) = %d\n",RX_remsize,rcc);
		return -1;
	}

	RX_rdTotal += rcc;

	if( RX_cachefp ){
		wcc = fwrite(buff,1,rcc,RX_cachefp);
		sendDistribution(Conn,RX_cachefp,RX_fsp,RX_tcp,buff,rcc);
		RX_binLen += wcc;

		if( wcc == 0 || RX_rdContLen+0x100000 < RX_binLen ){
			sv1log("discard cache: wcc=%d leng=%d/%d %s\n",
				wcc,RX_binLen,RX_rdContLen,RX_cpath);
			Ftruncate(RX_cachefp,0,0);
			fseek(RX_cachefp,0,0);
			RX_cachefp = NULL;
		}
	}

	if( !ClientEOF ){
	    wcc = fwriteResponseChunk(Conn,RX,buff,rcc);
	    RX_wrTotal += wcc;
	    if( wcc == 0 ){
		/* don't break to relay response data into the cache */
 sv1log("## HTTP out to client TIMEOUT[%d]\n",fileno(RX_tcp));
		/*fcloseTIMEOUT(RX_tcp);*/
		setClientEOF(Conn,RX_tcp,"binary flush-1");
	    }
	    if( wcc < rcc ){
		if( !ClientEOF )
			setClientEOF(Conn,RX_tcp,"binary flush-2");
 sv1log("## HTTP ERROR incomplete fwrite(%d)=%d %d/%d, SIGPIPE=%d\n",
 rcc,wcc,RX_rdTotal,RX_rdContLen,nsigPIPE);
		if( RX_tcp == RX_respBuff )
			abortHTTP("disk full?");

		if( RX_cachefp == NULL ){
 sv1log("## stop relay-2: FWRITE(%d)=%d, SIGPIPE=%d, no cache & no recipient\n",
 rcc,wcc,nsigPIPE);
			return -1;
		}
	    }
	    RX_noutput++;

	    Verbose("read %d / %d bytes / %d blocks; %d flush.\n",
		rcc,RX_rdTotal,RX_ninput,RX_nflush);
	}
	return wcc;
}

static distFromCacheSetup(Conn,RX,cpath,cachefp)
	Connection *Conn;
	ResponseContext *RX;
	char *cpath;
	FILE *cachefp;
{	char scdate[256];
	int lastmod;
	char *dcpath;

	lock_sharedNB(fileno(cachefp));
	stopDistribution(Conn,cachefp,cpath);

/*
 * copying too large file (larger than 1M bytes ?) should be avoided not to
 * make slower response to ther client and headvy load the the server...
 */

	dcpath = RX_dcx.d_cpath;
	RX_dcx.d_cpath[0] = 0;
	RX_dcx.d_fp = NULL;
	RX_dcx.d_ready = 0;
	RX_dcx.d_start = time(0);

	lastmod = HTTP_getLastModInCache(scdate,sizeof(scdate),cachefp,cpath);
	if( lastmod ){
		sprintf(dcpath,"%s#%d",cpath,lastmod);
		RX_dcx.d_fp = fopen(dcpath,"r");
		if( RX_dcx.d_fp != NULL ){
			lock_unlock(fileno(cachefp));
			Verbose("distFromCache: share %s\n",dcpath);
			RX_dcx.d_ready = 1;
		}else{
			RX_dcx.d_fp = dirfopen("distFromCache",dcpath,"w+");
			Verbose("distFromCache: created %s\n",dcpath);
		}
	}
	if( RX_dcx.d_fp == NULL ){
		RX_dcx.d_fp = TMPFILE("distFromCache");
		if( RX_dcx.d_fp == NULL ){
			RX_dcx.d_fp = fdopen(dup(fileno(cachefp)),"r");
			Verbose("distFromCache: WITHOUT-LOCK %s\n",cpath);
			RX_dcx.d_ready = -1;
		}
	}
	if( RX_dcx.d_ready == 0 ){
		copyfile1(cachefp,RX_dcx.d_fp);
		fflush(RX_dcx.d_fp);
		fseek(RX_dcx.d_fp,0,0);
		lock_unlock(fileno(cachefp));
	}
}
static distFromCacheDone(Conn,RX)
	Connection *Conn;
	ResponseContext *RX;
{
	if( RX_dcx.d_ready )
		lock_unlock(fileno(RX_cachefp));
	lock_unlock(fileno(RX_dcx.d_fp));
	fclose(RX_dcx.d_fp);

	if( RX_dcx.d_cpath[0] ){
		int ucode;
		ucode = unlink(RX_dcx.d_cpath);
		Verbose("dcxFromCache: %d unlink=%d [%ds] %s\n",
			RX_dcx.d_ready,ucode,
			time(0)-RX_dcx.d_start,RX_dcx.d_cpath);
	}
}
static cacheReadOK(cachefp)
	FILE *cachefp;
{
	if( cachefp == NULL )
		return 0;
	if( file_size(fileno(cachefp)) <= 0 )
		return 0;
	return 1;
}
static logCookie(Conn,RX,resCookie)
	Connection *Conn;
	ResponseContext *RX;
	char *resCookie;
{	char reqCookie[URLSZ];

	if( withCookie || *resCookie ){
		getFV(REQ_FIELDS,"Cookie",reqCookie);
		sv1log("[Cookie:%s][Set-Cookie:%s][Cache-Control:%s]\n",
			reqCookie,resCookie,RX_cachecontrol);
	}
}
static decomp_http_status(stat,resx)
	char *stat;
	HttpResponse *resx;
{	char *sp;
	int rcode;

	resx->hr_ver[0] = 0;
	resx->hr_reason[0] = 0;

	if( !STRH(stat,F_HTTPVER) ){
		rcode = 1200;
	}else{
		sp = wordScan(stat,resx->hr_ver);
		if( *sp == ' ' ) sp++;
		if( sp = scanint(sp,&rcode) ){
			lineScan(sp,resx->hr_reason);
			if( rcode <= 0 )
				rcode = R_BROKEN_RESPONSE;
		}else{
			sv1log("ERROR: BROKEN STATUS LINE: %s",stat);
			rcode = R_BROKEN_RESPONSE;
		}
	}
EXIT:
	resx->hr_rcode = rcode;
	return rcode;
}

static Fpurge(tc)
	FILE *tc;
{	int cfd,sfd;
	FILE *tmp;

	if( tc == NULL )
		return;

	tmp = fopen("/dev/null","w");
	if( tmp == NULL )
		tmp = tmpfile();
	if( tmp == NULL )
		return;
	cfd = fileno(tc);
	sfd = dup(cfd);
	dup2(fileno(tmp),cfd);
	fclose(tmp);
	fflush(tc);
	dup2(sfd,cfd);
	close(sfd);
	Ftruncate(tc,0,0);
}

extern char *TIMEFORM_mdHMS;
#define CMAP_APPROVER "Approver"
static int modwatch_enable;
static char *modwatch_notify;
static int modwatch_approver;

/*
 * HTTPCONF="modwatch:notifyto,approver" equals to
 * HTTPCONF="modwatch:notifyto=${ADMIN},approver=*:.localnet"
 */
static modwatch1(spec,Conn)
	char *spec;
	Connection *Conn;
{	char name[64],val[64],map[256];

	name[0] = val[0] = 0;
	sscanf(spec,"%[^=]=%s",name,val);
	sv1log("HTTPCONF=modwatch:%s=%s\n",name,val);
	if( strcmp(name,"notifyto") == 0 ){
		modwatch_notify = strdup(val);
		/* should verify the address here ... */
	}else
	if( strcmp(name,"approver") == 0 ){
		modwatch_approver = 1;
		if( val[0] == 0 )
			sprintf(map,"approve:*:%s","*:.localnet");
		else	sprintf(map,"approve:*:%s",val);
		scan_CMAP2(Conn,CMAP_APPROVER,map);
	}
	return 0;
}
HTTP_setModwatch(Conn,spec)
	Connection *Conn;
	char *spec;
{
	modwatch_enable = 1;
	scan_commaListL(spec,0,modwatch1,Conn);
}
static modwatchHead(Conn,RX)
	Connection *Conn;
	ResponseContext *RX;
{	char md5[64],mdpath[URLSZ],appath[URLSZ],approver[URLSZ],acap[128];
	char old[URLSZ],new[URLSZ],*cp,msg[URLSZ];
	char lmtime[64],event[URLSZ],*ep,*op;
	char omtime[64],ctime[64];
	char lpath[URLSZ];
	char url[URLSZ],req[URLSZ];
	FILE *afp,*log;
	int rcc,len,osize,oapproved;


	HTTP_originalURLx(Conn,url,sizeof(url));
	sprintf(req,"%s/%s %s",REQ_METHOD,REQ_VER,url);
	/* should use REQ_URL + Host: filed ? */
	toMD5(req,md5);
	Strins(md5+2,"/");

	sprintf(appath,"${ADMDIR}/approved/%s",md5);
	DELEGATE_substfile(appath,"",NULL,NULL,NULL);
	sprintf(mdpath,"${ADMDIR}/modified/%s",md5);
	DELEGATE_substfile(mdpath,"",NULL,NULL,NULL);

	afp = dirfopen("modwatchHead",appath,"r+");
	if( afp != NULL ){
		oapproved = 1;
	}else{
		oapproved = 0;
		afp = dirfopen("modwatchHead",mdpath,"r+");
	}
	if( afp != NULL ){
		rcc = fread(old,1,sizeof(old),afp);
		old[rcc] = 0;
		fseek(afp,0,0);
	}else{
		afp = dirfopen("modwatchHead",mdpath,"w");
		if( afp == NULL )
			return 0;
		rcc = 0;
		old[rcc] = 0;
	}
	StrftimeLocal(lmtime,sizeof(lmtime),TIMEFORM_mdHMS,RX_lastModTime);
	for( op = lmtime; *op; op++ )
		if( *op == ' ' )
			*op = '-';

	new[0] = 0;
	cp = new;
	cp += strlen(cp); sprintf(cp,"RQ=\"%s/%s %s\";",REQ_METHOD,REQ_VER,url);
	cp += strlen(cp); sprintf(cp,"CT=%s;",RX_ctype);
	cp += strlen(cp); sprintf(cp,"CL=%d;",RX_rdContLen);
	cp += strlen(cp); sprintf(cp,"MT=%s;",lmtime);
	cp += strlen(cp);
	len = cp - new;

	makeClientLog(Conn,approver);
	StrftimeLocal(ctime,sizeof(ctime),TIMEFORM_mdHMS,time(0));
	cp += strlen(cp); sprintf(cp,"AT=%s;",ctime);
	cp += strlen(cp); sprintf(cp,"AU=%s;",approver);

	if(  strncmp(old,new,len) == 0 ){
		fclose(afp);
		return 0;
	}

	sprintf(lpath,"${ADMDIR}/modified.log");
	DELEGATE_substfile(lpath,"",NULL,NULL,NULL);
	log = dirfopen("modified.log",lpath,"a");

	if( strstr(old+len,"DT=") == NULL ){
		/* the first detection of modification */
		if( old[0] )
			fseek(afp,-1,2);
		fprintf(afp,"DT=%s;\n",ctime);
		Ftruncate(afp,0,1);
		fseek(afp,0,0);

		ep = event;
		if( old[0] )
			sprintf(event,"modified: ");
		else	sprintf(event,"detected: ");
		ep += strlen(ep); sprintf(ep,"%s [%s]%d",url,lmtime,
			RX_rdContLen);
		ep += strlen(ep);
		if( old[0] ){
			osize = 0;
			if( op = strstr(old,"CL=") )
				sscanf(op,"CL=%d",&osize);
			omtime[0] = 0;
			if( op = strstr(old,"MT=") )
				sscanf(op,"MT=%[^;]",omtime);
			sprintf(ep," %d[%s]",osize,omtime);
		}
		sprintf(msg,"\tFILE:%s\n\tWHAT:%s\n\tOLD:%s\tNEW:%s\n",
			mdpath,event,old[0]?old:"\n",new);
		if( log ){
			fseek(log,0,2);
			fprintf(log,"-%s\n%s",event,msg);
			fflush(log);
		}
		if( modwatch_notify ){
			for( op = msg; *op; op++ )
				if( *op == '\t' )
					*op = ' ';
			notify_ADMINX(Conn,modwatch_notify,event,msg);
		}
	}

	if( modwatch_approver == 0 ){
		if( log != NULL )
			fclose(log);
		goto APPROVED;
	}

	ep = event;
	sprintf(event,"approved: ");
	ep += strlen(ep); sprintf(ep,"%s [%s]%d",url,lmtime,RX_rdContLen);
	ep += strlen(ep); sprintf(ep," %s",old[0]?"M":"D");
	ep += strlen(ep); sprintf(ep," [%s][%d]",approver,getpid());
	sprintf(msg,"\tFILE:%s\n\tWHAT:%s\n\tOLD:%s\tNEW:%s\n",
		mdpath,event,old[0]?old:"\n",new);

	if( old[0] )
	if( find_CMAP(Conn,CMAP_APPROVER,acap) < 0
		 /* the client is not an approver */
	 /* && age < 1week ? */
	){
		Fpurge(RX_tc_sav);
		Fpurge(RX_tcp);
		fprintf(RX_tcp,"HTTP/%s 503 Service Unavailable\r\n",
			MY_HTTPVER);
		fprintf(RX_tcp,"\r\n");
		fprintf(RX_tcp,"Service Unavailable\r\n");
		fflush(RX_tcp);
		RespCode = RX_code = 503;

		if( oapproved ){
			renameRX(appath,mdpath);
		}
		if( log ){
			fseek(log,0,2);
			fprintf(log,"-un%s\n",event);
			fclose(log);
		}
		fclose(afp);
		return -1;
	}
	if( log ){
		fseek(log,0,2);
		fprintf(log,"-%s\n%s",event,msg);
		fclose(log);
	}
	if( modwatch_notify ){
		for( op = msg; *op; op++ )
			if( *op == '\t' )
				*op = ' ';
		notify_ADMINX(Conn,modwatch_notify,event,msg);
	}
	if( !oapproved ){
		renameRX(mdpath,appath);
	}
	sv1tlog("modwatch-old: %s",old);
	sv1tlog("modwatch-new: %s",new);

APPROVED:
	fprintf(afp,"%s\n",new);
	Ftruncate(afp,0,1);
	fclose(afp);
	return 0;
}

static checkServ(Conn)
	Connection *Conn;
{
	if( 0 < CheckServ ) return 1;
	if( CheckServ < 0 ) return 0;

	if( !HTTP_reqWithHeader(Conn,REQ) ){
		CheckServ = -1;
		return 0;
	}
	if( 1024 <= DST_PORT ){
		CheckServ = -1;
		return 0;
	}
	if( DST_PORT == 80 || DST_PORT == 443 )
	if( BadRequest == 0 && Normalized == 0 ){
		CheckServ = -1;
		return 0;
	}

	CheckServ = 1;
	return 1;
}
static badServ(Conn,where,fs,statline)
	Connection *Conn;
	char *where;
	FILE *fs;
	char *statline;
{	char buf[128];
	int rcc,timeout;

	if( 0 < BadServ ) return 1;
	if( BadServ < 0 ) return 0;

	if( !checkServ(Conn) )
		return 0;

	if( HTTP_WAIT_BADSERV < WaitServ )
		return 0;

	if( statline != NULL ){
		rcc = strlen(statline);
		goto check;
	}

	if( file_isreg(fileno(fs)) ){ /* maybe NULLFP() */
		BadServ = -1;
		return 0;
	}

	if( BadRequest || Normalized ){
		if( DST_PORT == 80 || DST_PORT == 443 )
			timeout = 100;
		else	timeout = 500;
	}else{
		if( DST_PORT == 80 || DST_PORT == 443 )
			timeout = 1;
		else	timeout = 10;
	}

	WaitServ += timeout;
	if( fPollIn(fs,timeout) <= 0 )
		return 0;
	rcc = fgetBuffered(buf,sizeof(buf),fs);
	if( rcc == 0 )
		rcc = recvPeekTIMEOUT(fileno(fs),buf,sizeof(buf)-1);
	if( rcc <= 0 )
		return 0;
	buf[rcc] = 0;
	statline = buf;
	sv1log("## badServer? %s: BQ=%d+%d %d: %s",
		where,BadRequest,Normalized,rcc,statline);

check:
	if( strncmp(statline,"HTTP/",5) == 0 ){
		BadServ = -1;
		return 0;
	}
	sv1log("## badServer! %s: BQ=%d+%d %d: %s",
		where,BadRequest,Normalized,rcc,statline);
	lineScan(where,BadServDetected);
	lineScan(statline,BadServResponse);
	httpStat = CS_BADREQUEST;
	BadServ = 1;
	return 1;
}
static putBadRequest(Conn,tc)
	Connection *Conn;
	FILE *tc;
{
	sv1log("## badRequest/Server: Server[%s/%dms][%s] Request[%d+%d][%s]\n",
		BadServDetected,WaitServ,
		BadServResponse,BadRequest,Normalized,OREQ_MSG);

	fprintf(tc,"HTTP/%s 400 Bad Request\r\n",MY_HTTPVER);
	fprintf(tc,"Connection: close\r\n");
	fprintf(tc,"\r\n");
	fprintf(tc,"%s",OREQ_MSG);
	http_Log(Conn,400,CS_BADREQUEST,REQ,0);
}

static relay_response(Conn,cpid,proto,server,iport,req,acpath,fromcache,
afs,atc,afc,acachefp,cache_rdok)
	Connection *Conn;
	char *proto;
	char *server;
	char *req;
	char *acpath;
	FILE *afs;
	FILE *atc;
	FILE *afc;
	FILE *acachefp;
{	ResponseContext RXbuf, *RX = &RXbuf;
	Referer refererb;
	char refbuf[URLSZ];
	void (*osigpipe)();
	int lastRcc;
	char *line,*got1;
	int rcc;

	RX_Conn = Conn;
	RX_tmpFileId = "HTTP-respBuff";
	RX_cachefp = acachefp;
	RX_cpath = acpath;
	RX_referer = &refererb;
	RX_inHeader = 1;

	RX_tc_sav = 0;
	RX_tcp = atc;
	RX_fsp = afs;
	RX_fsx.t_EOF = 0;
	RX_fsx.t_lastRcc = 0;
	RX_fsx.t_chunked = 0;
	RX_fsx.t_chunk_ser = 0;
	RX_fsx.t_chunk_siz = 0;
	RX_fsx.t_chunk_rem = 0;
	RX_fsx.t_keepAlive = 0;
	RX_remsize = sizeof(RX_rdBuff) - 1;

	line = RX_rdBuff;

	if( Conn->xf_filters & (XF_FCL|XF_FTOCL) )
		RX_tcx.t_chunked = 0;
	else	RX_tcx.t_chunked = ClntAccChunk;
	RX_tcx.t_chunk_ser = 0;

	RX_Start = Time();
	RX_qWithHeader = HTTP_reqWithHeader(Conn,req);
	wordScan(req,RX_qmethod);
	RX_reqHEAD = strcasecmp(RX_qmethod,"HEAD") == 0;

	RX_tryKeepAlive = 0;
	if( ClntKeepAlive ){
		if( strcaseeq(RX_qmethod,"GET")||strcaseeq(RX_qmethod,"HEAD") )
			RX_tryKeepAlive = 1;
		else
		clntClose(Conn,"M:Method is ``%s'' cpid=[%d]",RX_qmethod,cpid);
	}

	for(;;){

	got1 = GETRESPTXT(Conn,RX,fromcache);
	if( badServ(Conn,"RESP",afs,line) ){
		return R_EMPTY_RESPONSE;
	}
	lastRcc = RX_fsx.t_lastRcc;
	RespCode = decomp_http_status(line,&RX_status);
	RX_Resp1 = Time();

		if( RX_code == 100 ){
			sv1log("#HT11 [%s] Skip %s",REQ_VER,line);
			if( 0 <= vercmp(REQ_VER,"1.1") )
				fputs(line,RX_tcp);
			while( got1 = GETRESPTXT(Conn,RX,fromcache) ){
				if( 0 <= vercmp(REQ_VER,"1.1") )
					fputs(line,RX_tcp);
				sv1log("#HT11 [%s] Skip %s",REQ_VER,line);
				if( line[0] == '\r' || line[0] == '\n' )
					break;
			}
			if( feof(RX_fsp) ){
				sv1log("#HT11 Skip ... EOF\n");
			}else{
				fflush(RX_tcp);
				continue;
			}
		}
		break;
	}

	if( RX_code != 1200 )
	if( RX_tcx.t_chunked ){
		if( strcmp(RX_ver,"HTTP/1.1") < 0 ){ 
			sv1log("#HT11 resp version: %s -> %s\n",RX_ver,"1.1");
			sprintf(line,"HTTP/1.1 %d %s\r\n",RX_code,RX_reason);
			lastRcc = RX_fsx.t_lastRcc = strlen(line);
		}
	}

	if( cpid )
		Kill(cpid,SIGTERM);
	if( got1 == NULL ){
		sv1log("HTTP realy_response: EOF at start\n");
		http_Log(Conn,500,CS_EOF,req,0);
		return R_EMPTY_RESPONSE;
	}

	if( RX_cachefp != NULL && 500 <= RX_code ){
		if( cache_rdok ){
			sv1log("HTTP temporary error ? %s",line);
			http_Log(Conn,RX_code,CS_EOF,req,0);
			copy_fileTIMEOUT(RX_fsp,NULLFP(),NULL);
			return R_EMPTY_RESPONSE;
		}
		RX_cachefp = NULL;
		DontTruncate = 1;
	}

	if( RX_cachefp != NULL ){
		Verbose("CacheMtime: %d : %d\n",CacheMtime,
			file_mtime(fileno(RX_cachefp)));

		if( CacheMtime != file_mtime(fileno(RX_cachefp)) ){
			int fd;

			sv1log("## !!! Cache is updated by someone else.\n");
/*
RX_cachefp = NULL;
*/
			fd = open(RX_cpath,0);
			dup2(fd,fileno(RX_cachefp));
			close(fd);
			DontTruncate = 1;
		}
	}

	/*
	 *	If the transmission speed from the server is slow,
	 *	do flush frequently.
	 */
	if( fromcache ){
		RX_connDelay = 0;
	}else{
		int delay;

		RX_connDelay = CONN_DONE - CONN_START;
		RX_firstResp = Time() - CONN_DONE;
		daemonlog("D","connDelay:%5.2fsec, firstResp:%5.2fsec\n",
			RX_connDelay,RX_firstResp);
	}

	if( checkClientEOF(Conn,RX_tcp,"start_response") ){
		httpStat = CS_EOF;
		return R_UNSATISFIED;
	}

	RX_noBody = RX_reqHEAD;
	RX_errori = 0;
	RX_isText = 0;
	RX_rdHeadTotal = 0;
	RX_rdTotal = 0;
	RX_wrTotal = 0;
	RX_wrHeadTotal = 0;
	RX_wrbufed = 0;
	RX_putPRE = 0;
	RX_isHTTP09 = 0;
	RX_lastMod = 0;
	RX_lastModTime = 0;
	RX_setCookie[0] = 0;
	RX_cachecontrol[0] = 0;

	RX_ctype[0] = 0;
	RX_connection[0] = 0;
	RX_cencoding[0] = 0;
	RX_servername[0] = 0;
	RX_tencoding[0] = 0;
	RX_rdContLen = 0;
	RX_txtLen = 0;
	RX_binLen = 0;
	RX_bodyLines = 0;
	RX_nflush = 0;
	RX_noutput = 0;
	RX_niced = 0;
	RX_respBuff = NULL;

	if( RX_code == 403 )
		HTTP_delayReject(Conn,req,line,0);
	else
	if( RX_code == 404 )
		delayUnknown(Conn,DO_DELEGATE||IsMounted,OREQ);

/* This will not work well when the client is DeleGate with cache
 * and  its cache is newer than the cache of this DeleGate.
 */
	if( RX_code == 304 ){
		int cacheOk;
		int forceBody;

		cacheOk = cacheReadOK(RX_cachefp);
		sv1log("HTTP status: %d %s => %x/%d\n",
			RX_code,RX_reason,RX_cachefp,cacheOk);

		if( cacheOk ){
			ftouch(RX_cachefp,time(0));
			forceBody = forceBodyResponse(Conn);
		}else{
			forceBody = 0;
			sv1log("No-Cache to reuse\n");
		}

		if( ClntIfMod[0] && !forceBody && CacheLastMod<=ClntIfModClock){
			Verbose("%d is relayed to the client.\n",RX_code);
			if( RX_cachefp ){
				Verbose("DontWriteCache 304 not modified.\n");
				/*fclose(RX_cachefp);*/
				RX_cachefp = NULL;
				DontTruncate = 2;
			}
		}else
		if( RX_cachefp ){
/*
 * 304 should be passed thru to the client if it is a delegated
 * with cache file which is very same with this delegated.
 */
			sv1log("HTTP <=+= %d [%s] %s",RespCode,RX_cpath,req);
			httpStat = CS_STABLE;

			distFromCacheSetup(Conn,RX,RX_cpath,RX_cachefp);
			RX_code = relay_response(Conn,0,proto,server,iport,req,
				RX_dcx.d_cpath,1,RX_dcx.d_fp,RX_tcp,NULL,NULL,1);
			distFromCacheDone(Conn,RX);
			return RX_code;
		}
		RX_noBody = 1;
		RX_lastMod = 1;
	}else{
		if( DontReadCache ){
			httpStatX = CS_RELOAD;
		}else
		if( RX_cachefp != NULL && cache_rdok )
			httpStat = CS_OBSOLETE;
	}

	if( RX_code == 1200 ){
		char xline[32],*np;

		RX_rdTotal = lastRcc;
		RX_isHTTP09 = 1;
		RX_inHeader = 0;

		if( strchr(line,'\n') )
			RX_isText = TX_HTML; /* HTML is assumed */

		Strncpy(xline,line,sizeof(xline));
		if( strtailchr(xline) != '\n' )
			strcat(xline," ...\n");

		sv1log("No header (HTTP/0.9) : %s assumed : [%x][%x]%s",
			(RX_isText?"text":"binary"), xline[0],xline[1],xline);
	}else{
		RX_rdHeadTotal = lastRcc;
		switch( RX_code ){
		case 200:
		case 301:
		case 302: /* OK and cachable */
			break;
		case 304:
			break;
		case 305:
		case 306:
			break;
		case 206:
			if( RX_cachefp == NULL )
				break;
			sv1log("-- 206 is NOT error but no-cache ...\n");
		default:
			sv1log("HTTP error request: %s",req);
			sv1log("HTTP error status: %d %s\n",RX_code,RX_reason);
			RX_errori = -RX_code;
		}
	}

	setReferer(Conn,proto,server,iport,req,RX_referer,refbuf);

	RX_convChar = cur_codeconvCL(NULL) || CCXactive(CCX_TOCL);
	RX_deEnt = toProxy && protoGopher(DST_PROTO);

	if( IsGopherSearch && RX_isHTTP09 )
		sv1log("Gopher/HTTP relay skipped 1st response line\n");
	else
	if( (RX_code==305 || RX_code==306) && (ConnType=='p' || ConnType=='m') ){
		RESP_SAV = 1;
		/* RX_tcp = NULLFPW(); */
		RX_tcp = TMPFILE("RESP_MSG");
		strcpy(RESP_MSG,line);
		RESP_LEN = strlen(line);
	}else{
		if( !RX_inHeader && RX_isText==TX_HTML ){
			RX_wrTotal += FPUTS(line,"text/html",!RX_inHeader && RX_convChar,RX_deEnt);
		}else{
			if( RX_qWithHeader || !RX_inHeader )
				RX_wrHeadTotal += fwrite(line,1,lastRcc,RX_tcp);
		}

		if( !RX_isHTTP09 && RX_inHeader ){
			if( !ImResponseFilter )
			if( RX_qWithHeader )
				HTTP_putDeleGateHeader(Conn,RX_tcp,fromcache);
				/* also X-Request-XXX is echoed */
		}

		if( RX_cachefp ){
			fwrite(line,1,lastRcc,RX_cachefp);
			sendDistribution(Conn,RX_cachefp,RX_fsp,RX_tcp,line,lastRcc);
		}
	}


	if( !ClientEOF )
	if( 0 < DELEGATE_LINGER )
		set_linger(fileno(RX_tcp),DELEGATE_LINGER);
	if( fromcache )
		RX_fsd = -1;
	else	RX_fsd = fileno(RX_fsp);
	RX_tcd = fileno(RX_tcp);

	if( WillKeepAlive ){
		if( CKA_RemAlive <= 1 )
			clntClose(Conn,"f:finished lifetime");
		else
		if( RX_errori != 0 )
			clntClose(Conn,"s:bad status: %d",RX_errori);
	}

	if( RX_code == 206 ){
		sv1log("#HT11 NO-response-buffering: 206 Partial\n");
	}else
	if( RX_tcx.t_chunked ){
		sv1log("#HT11 NO-response-buffering: chunked mode\n");
	}else
	if( !ClientEOF )
	if( !RX_noBody && !RX_isHTTP09 && RX_qWithHeader )
	if( withConversion(Conn,0)
	 || WillKeepAlive && (!fromcache || fromcache && !HTTP_ContLengOk(RX_fsp)) ){
		RX_respBuff = getTmpFile(RX_tmpFileId,Conn);
		RX_tc_sav = RX_tcp;
		RX_tcp = RX_respBuff;
		RX_tcd = -1;
	}

	RX_nsigreport = 0;
	nsigPIPE = 0;
	osigpipe = Vsignal(SIGPIPE,sigPIPE);
	Where = "header";

	for( RX_ninput = 0;;RX_ninput++ ){
	    if( !RX_inHeader && RX_noBody ){
		sv1log("#HT11 NO-BODY: remsize=%d\n",RX_remsize);
		break;
	    }
	    if( RX_fsx.t_EOF )
		break;
	    if( feof(RX_fsp) )
		break;
	    if( nsigPIPE ){
		if( !ClientEOF )
			setClientEOF(Conn,RX_tcp,"relay_response.SIGPIPE");
		if( RX_nsigreport++ % 100 == 0 ){
			sv1log(
"## (%d) SIGPIPE * %d: the recipient seems to be dead, Ctype:%s\n",
			RX_nsigreport,nsigPIPE, RX_ctype);
		}
	    }
	    if( ClientEOF && RX_cachefp == NULL ){
 sv1log("## stop relay-1: no recipient & no cache\n");
		break;
	    }
	    if( ClientEOF && RX_cachefp != NULL && CACHE_TAKEOVER ){
		if( CACHE_TAKEOVER < time(0) - RX_Start ){
			sv1log("CACHE_TAKEOVER: TIMEOUT=takeover:%d\n",
				time(0)-(int)RX_Start);
			break;
		}
	    }

	    if( 10 < RX_ninput )
		RX_niced = doNice("HTTPresponse",RX_fsd,RX_tcd,RX_niced,
RX_rdTotal,RX_ninput,RX_Resp1);

 	/*
 	 * stop buffering for Keep-Alive when the response data is large
 	 * or response speed is slow.
 	 */
	RX_fsx.t_nready = 0;

         if( WillKeepAlive )
         if( RX_tcp == RX_respBuff && !RX_inHeader && !ClientEOF )
         if( 1024*1024 < RX_rdContLen
	 || 1024*1024 < RX_rdTotal
	 || 15 < Time()-CONN_DONE
 	 || !fromcache
	 && RX_fsx.t_nready == 0 && (RX_fsx.t_nready = fPollIn(RX_fsp,100)) == 0
	 && 10 < Time()-CONN_DONE )
                 RX_tcp = detach_respbuff(Conn,RX_tcp,RX_tc_sav,
                         RX_qWithHeader,"flush slow response.");
	/*
	if( RX_fsx.t_nready < 0 ){
		sv1log("Polling server response failed, maybe EOF\n");
		break;
	}
	*/

	    if( !RX_inHeader && RX_fsx.t_chunked ){
		lastRcc = RX_fsx.t_lastRcc;
		if( RX_fsx.t_chunk_ser && 0 < lastRcc ){
			Verbose("#HT11 --getChunk[%d] data  : %d / %d / %d\n",
				RX_fsx.t_chunk_ser,lastRcc,
				RX_fsx.t_chunk_rem,RX_fsx.t_chunk_siz);

			RX_fsx.t_chunk_rem -= lastRcc;
			RX_remsize = RX_fsx.t_chunk_rem;
			if( sizeof(RX_rdBuff)-1 < RX_remsize )
				RX_remsize = sizeof(RX_rdBuff) -1;
		}
		if( RX_fsx.t_chunk_rem == 0 ){
			if( getChunk(RX) < 0 )
				break;
			RX_fsx.t_lastRcc = 0;
			continue;
		}
	    }else
	    if( !RX_inHeader && RX_fsx.t_keepAlive && 0 < RX_rdContLen ){
		RX_remsize = RX_rdContLen - RX_rdTotal;
		if( sizeof(RX_rdBuff)-1 < RX_remsize )
			RX_remsize = sizeof(RX_rdBuff)-1;
		Verbose("#HT11 --Length=%d = %d + %d\n",
			RX_rdContLen,RX_rdTotal,RX_remsize);
		if( RX_rdContLen <= RX_rdTotal )
			break;
	    }
 
	    if( RX_inHeader || RX_isText ){
		if( !RX_inHeader && RX_bodyLines == 0 ){
			RX_convChar = cur_codeconvCL(NULL)||CCXactive(CCX_TOCL);
			check_codeconv(1);
			if( !ClientEOF && RX_putPRE == 1 ){
				RX_wrTotal += FPUTS("\n<PRE>",RX_ctype,
					RX_convChar,0);
				RX_putPRE = 2;
			}
		}
		if( !RX_inHeader && !fromcache && !ClientEOF
		 && RX_fsx.t_nready == 0 && (RX_fsx.t_nready = fPollIn(RX_fsp,200)) == 0 ){
			RX_noutput++;
			if( fflushTIMEOUT(RX_tcp) == EOF )
				setClientEOF(Conn,RX_tcp,"text flush-1");
			else
			if( RX_rdContLen == 0 && RX_tcp == RX_respBuff ){
				RX_tcp = detach_respbuff(Conn,RX_tcp,RX_tc_sav,
					RX_qWithHeader,"flush endless text.");
			}
		}

		if( GETRESPTXT(Conn,RX,fromcache) == NULL ){
textEOF:
			if( !ClientEOF && !RX_inHeader )
			if( !RX_noBody )
				RX_wrTotal += FPUTS("",RX_ctype,RX_convChar,0);
			break;
		}
		lastRcc = RX_fsx.t_lastRcc;

		if( RX_inHeader && (line[0] == '\r' || line[0] == '\n') )
			endRespHead(Conn,RX);

		if( !RX_inHeader && RX_isText && RX_isBin ){ /* RX_isBin is set in GETRESPTXT() */
			sv1log("non-text in text type resp.(%d/%d)\n",RX_rdTotal,RX_ninput);
			RX_errori = R_NONTEXT_INTEXT;
			rcc = lastRcc;
			Where = "binaryBody";
			RX_isText = 0;
			goto GOT_BINARY;
		}

		/* output to the cache should be
		 * before "FPUTS()" which may rewrite input line
		 * after "goto GOT_BINARY" to cache binary line with fwrite()
		 */
		if( RX_cachefp && !RX_isHTTP09 ){
			if( !(RX_inHeader && STRH(line,F_DGVer)) )
			if( !(RX_inHeader && STRH(line,F_TransEncode)) )
			if( !(RX_inHeader && STRH_Connection(line)) ){
				fputs(line,RX_cachefp);
				sendDistribution(Conn,RX_cachefp,RX_fsp,RX_tcp,line,lastRcc);
			}
			if( !RX_inHeader )
				RX_txtLen += lastRcc;
		}

		if( !RX_inHeader ){
			RX_rdTotal += lastRcc;
			RX_bodyLines += 1;

			/* don't cause automatic flush at
			 * non line boundary.  */
			if( OBUFSIZE <= RX_wrbufed+lastRcc
			|| !fromcache && READYCC(RX_fsp) == 0 ){ 
				if( !ClientEOF && fflushTIMEOUT(RX_tcp) == EOF )
					setClientEOF(Conn,RX_tcp,"text flush-2");
				RX_noutput++;
				RX_wrbufed = lastRcc;
			}else{
				RX_wrbufed += lastRcc;
				/* should be length of converted string. */
			}
		}

		relayTxtResp1(Conn,RX,req);
		if( RX_inHeader ){
			RX_rdHeadTotal += lastRcc;
			scanRespHead1(Conn,RX,line);
			if( !RX_inHeader ){

 sv1log("%s %d Content-{Type:%s Encoding:[%s/%s] Leng:%d} Server:%s\n",
 RX_ver,RX_code,RX_ctype,RX_cencoding,RX_tencoding,RX_rdContLen,RX_servername);

				if( !RX_noBody )
				if( modwatch_enable )
				if( modwatchHead(Conn,RX) != 0 )
					break;

				Where = "headerFlush";
				if( !ClientEOF && RX_tcp != RX_respBuff )
					flushHead(Conn,RX_tcp,RX_rdContLen);
				RX_noutput++;

				if( RX_reqHEAD )
					break;

				if( RX_isText )
					Where = "textBody";
				else	Where = "binaryBody";
			}
		}
		if( RX_rdContLen+0x100000 < RX_txtLen ){
			if( RX_tcp == RX_respBuff ){
				RX_tcp = detach_respbuff(Conn,RX_tcp,RX_tc_sav,
					RX_qWithHeader,"huge data");
			}
			if( RX_cachefp ){
				sv1log("discard cache: leng=%d/%d %s\n",
					RX_txtLen,RX_rdContLen,RX_cpath);
				Ftruncate(RX_cachefp,0,0);
				fseek(RX_cachefp,0,0);
				RX_cachefp = NULL;
			}
		}
	    }else{
		if( RX_tcp == RX_respBuff && !WillKeepAlive ){
			RX_tcp = detach_respbuff(Conn,RX_tcp,RX_tc_sav,
				RX_qWithHeader,"non-text data, non keep-alive");
		}
		rcc = GETRESPBIN(Conn,RX,fromcache,RX_fsx.t_nready);
GOT_BINARY:
		if( relayBinResp1(Conn,RX,rcc) < 0 )
			break;
	    }
	}

	if( !RX_noBody ){
		putChunk(RX,"",0);
		Where = "bodyFlush";

		if( !(fromcache && ClientEOF) )
		if( RX_rdTotal < RX_rdContLen )
			RX_errori = R_UNSATISFIED;

		if( !ClientEOF ){
			if( RX_putPRE == 2 ){
			RX_wrTotal += FPUTS("</PRE>\n",RX_ctype,RX_convChar,0);
			}
		}
	}

	RX_wrContLen = RX_rdContLen;
	if( RX_tcp == RX_respBuff )
		releaseRespbuf(Conn,RX);

	if( RX_errori == R_NONTEXT_INTEXT )
		if( RX_rdTotal == RX_rdContLen )
			RX_errori = 0;

	if( WillKeepAlive )
		abortKeepAlive(Conn,RX);

	if( file_isreg(fileno(RX_tcp)) ){
		sv1log("DON'T CLOSE RESPONSE:(%d) %s\n",
			reqRobotsTxt(Conn),REQ_URL);
	}else
	if( RESP_SAV )
		fclose(RX_tcp);
	else
	if( checkClientEOF(Conn,RX_tcp,"flush_body") ){
		httpStat = CS_EOF;
	}else{
		Verbose("HTTP RESPONSE FLUSH: DO\n");
		if( WillKeepAlive ){
			fflushKeepAlive(Conn,"flush_response",
				afc,RX_tcp,RX_wrHeadTotal+RX_wrTotal);
			if( nsigPIPE || ClientEOF )
				clntClose(Conn,"p:premature EOF on flush");
			if( nsigPIPE && !ClientEOF )
				setClientEOF(Conn,RX_tcp,"flush_response.SIGPIPE");
			if( ClientEOF )
				httpStat = CS_EOF;
		}else{
			tcCLOSED = 1;
			if( fcloseTIMEOUT(RX_tcp) == EOF )
				httpStat = CS_EOF;
			setClientEOF(Conn,RX_tcp,NULL);
		}
	}

	if( RX_cachefp )
		fflush(RX_cachefp);

	Vsignal(SIGPIPE,osigpipe);

	http_log(Conn,proto,server,iport,req,RX_code,RX_ctype,RX_rdTotal,
		RX_lastModTime,CONN_DONE-CONN_START,Time()-CONN_DONE);

sprintf(line,"HTTP transmitted: %dhead+%d/%dbody=>%dtxt+%dbin->%d/%d, %di/%do/%df/%3.1f",
		RX_rdHeadTotal,RX_rdTotal,RX_rdContLen,
		RX_txtLen,RX_binLen,
		RX_wrTotal,RX_wrContLen,
		RX_ninput,RX_noutput,RX_nflush,
		Time(0)-RX_Start);
 	sv1log("%s\n",line);

	if( withCookie || *RX_setCookie )
		logCookie(Conn,RX,RX_setCookie);

	if( RX_errori == R_UNSATISFIED )
		return RX_errori;

	if( RX_lastMod == 0 ){
		if( RX_code == 302 ){
			/* relatively shorter expire time is desired ? */
		}else{
			Verbose("No Last-Modified:\n");
#ifndef CACHE_NOLM
			return R_GENERATED;
#endif
		}
	}

	return RX_errori;
}

static HTTP_relayCachedHeader(Conn,cache,tc,resx)
	Connection *Conn;
	FILE *cache,*tc;
	HttpResponse *resx;
{	char line[1024];
	int lines;
	int total;

	total = 0;
	for(lines = 0;;lines++){
		if( fgets(line,sizeof(line),cache) == NULL )
			break;
		if( lines == 0 ){
			if( strncasecmp(line,F_HTTPVER,F_HTTPVERLEN) == 0 ){
				decomp_http_status(line,resx);
				continue;
			}
		}
		if( strncasecmp(line,"Date:",5) == 0 ){
			strcpy(line,"Date: ");
			StrftimeGMT(line,sizeof(line),TIMEFORM_RFC822,time(0));
			strcat(line,"\r\n");
		}
		if( line[0] == '\r' || line[0] == '\n' )
			total += putKeepAlive(Conn,tc);

		total += strlen(line);
		fputs(line,tc);
		if( line[0] == '\r' || line[0] == '\n' )
			break;
	}
	return total;
}
static relay_response_fromCache(Conn,proto,server,iport,req,cpath,cachefp,tc,fc)
	Connection *Conn;
	char *proto,*server;
	char *req;
	char *cpath;
	FILE *cachefp,*tc,*fc;
{	int rcode;
	int mycache;
	char scdate[256];
	int rtotal;
	HttpResponse resx;

	httpStat = CS_HITCACHE;
	ConnType = 'c';

	if( reqRobotsTxt(Conn) ){
		rtotal = putRobotsTxt(Conn,tc,cachefp,1);
		http_log(Conn,proto,server,iport,req,200,"text/plain",rtotal,0,
			-1.0,Time()-CONN_DONE);
		return 0;
	}

	if( 0 < ClntIfModClock && !forceBodyResponse(Conn) ){
		mycache = HTTP_getLastMod(scdate,sizeof(scdate),cachefp,cpath);
		if( mycache <= ClntIfModClock ){
			char ctype[128];

			/* dont send the same or older resource */
			fprintf(tc,"HTTP/1.0 304 Not Modified (cached)\r\n");
			rtotal = HTTP_relayCachedHeader(Conn,cachefp,tc,&resx);
			flushHead(Conn,tc,0);
			sv1log("HTTP <=S= %d [%s] %s",resx.hr_rcode,cpath,req);

			ctype[0] = 0;
			http_log(Conn,proto,server,iport,req,304,ctype,rtotal,0,
				-1.0,Time()-CONN_DONE);
			return 0;
		}
	}
	rcode = relay_response(Conn,0,proto,server,iport,req,cpath,1,
		cachefp,tc,fc,NULL,1);

	sv1log("HTTP <=-= %d [%s] %s",RespCode,cpath,req);

	if( rcode < 0 || /*rcode == 304 ||*/ RespCode == 304 ){
		sv1log("----------- UNLINK:%d:%d:%s\n",rcode,RespCode,cpath);
		unlink(cpath);
		return rcode;
	}
	if( /*rcode == 304*/ RespCode == 304 ){
		sv1log("ERROR  304 Not modified messages was cached?\n");
		return 304;
	}
	return 0;
}


/*###################################### REQUEST */
typedef struct {
	Connection *q_Conn;
	int	q_cpid;
	int	q_totalc;
	int	q_rcode;
	int	q_tsfd;
	int	q_fsfd;

	char   *q_proto;
	char   *q_site;
	char	q_site_buf[256];
	int	q_port;
	char   *q_upath;

	int	q_useCache;
	int	q_expire;
	char	q_cpath[1024];
	FILE   *q_cachefp;
	int	q_cdate;
	int	q_cacheRemove;

	int	q_range[2];
} QueryContext;

#define QX_Conn		QX->q_Conn
#define QX_cpid		QX->q_cpid
#define QX_totalc	QX->q_totalc
#define QX_rcode	QX->q_rcode
#define QX_tsfd		QX->q_tsfd
#define QX_fsfd		QX->q_fsfd

#define QX_proto	QX->q_proto
#define QX_site		QX->q_site
#define QX_site_buf	QX->q_site_buf
#define QX_port		QX->q_port
#define QX_upath	QX->q_upath

#define QX_useCache	QX->q_useCache
#define QX_expire	QX->q_expire
#define QX_cpath	QX->q_cpath
#define QX_cachefp	QX->q_cachefp
#define QX_cdate	QX->q_cdate
#define QX_cacheRemove	QX->q_cacheRemove

#define QX_range	QX->q_range

editRequestHeader(Conn,fs)
	Connection *Conn;
	FILE *fs;
{
}
static relaydata(Conn,fc,ts,timeout)
	Connection *Conn;
	FILE *fc,*ts;
{	int fcd,rcc,bytes;
	char buf[0x2000];

	bytes = 0;
	while( rcc = fgetBuffered(buf,sizeof(buf),fc) ){
		Fwrite(buf,1,rcc,ts);
		if( PadCRLF && buf[rcc-1] != '\n' ){
			sv1log("added CRLF at endOfLine for Mozilla/0.9(^_^;\n");
			fputs("\r\n",ts);
		}

		fflush(ts);
		bytes += rcc;

		strncpy(OREQ_MSG+OREQ_LEN,buf,rcc+1);
		OREQ_LEN += rcc;

		sv1log("Relay_bufferd (%d bytes)\n",rcc);
	}
	fcd = fileno(fc);
	for(;;){
		if( (rcc = readTimeout(fcd,buf,sizeof(buf),timeout)) == 0 )
			break;
		if( Fwrite(buf,1,rcc,ts) <= 0 )
			break;
		fflush(ts);
		sv1log("Relay_nonbuffered (%d+%d bytes)\n",bytes,rcc);
		bytes += rcc;

		strncpy(OREQ_MSG+OREQ_LEN,buf,rcc+1);
		OREQ_LEN += rcc;
	}
	fflush(ts);
	return bytes;
}
static HTTP_repairRequest(req)
	char *req;
{	int nsp,repaired;
	char xreq[URLSZ],*xp,*sp,sc;

	nsp = 0;
	repaired = 0;
	xp = xreq;
	for( sp = req; sc = *sp; sp++ ){
		if( sc == ' ' || sc == '\t' ){
			nsp++;
			if( 2 <= nsp && strncmp(sp+1,F_HTTPVER,F_HTTPVERLEN) != 0 ){
				sprintf(xp,"%%%02x",sc);
				xp += strlen(xp);
				repaired++;
				continue;
			}
		}
		*xp++ = sc;
	}
	*xp = 0;
	if( repaired ){
		strcpy(req,xreq);
		sv1log("REQUEST REPAIRED: %s",xreq);
	}
	return repaired;
}

#define SPorHT(ch)	(ch == ' ' || ch == '\t')
#define CRorLF(ch)	(ch == '\r' || ch == '\n')
static char *fgets822FromC(Conn,buff,size,fc,normalize)
	Connection *Conn;
	char *buff;
	FILE *fc;
{	char *rcode,*rcode1,*bp,*hp;
	int rsize,ch;

	rcode = 0;
	bp = buff;
	rsize = size;
	while( 1 < rsize ){
		rcode1 = DDI_fgetsFromC(Conn,bp,rsize,fc);
		if( bp == buff )
			rcode = rcode1;
		if( rcode == NULL || !normalize )
			break;
		ch = buff[0];
		if( CRorLF(ch) )
			break;

		if( bp != buff && normalize ){
			for( hp = bp; ch = *hp; hp++ )
				if( !SPorHT(ch) )
					break;
			if( hp != bp + 1 ){
				sv1log("##HHu reduce leading spaces<<%s",bp);
				strcpy(bp+1,hp);
				Normalized++;
			}
		}
		if( strtailchr(bp) != '\n' )
			break;
		ch = DDI_peekcFromC(Conn,fc);
		if( ch == EOF )
			break;
		if( !SPorHT(ch) )
			break;

		sv1log("##HHu unfold header<<%s",bp);
		hp = bp;
		for( bp += strlen(bp) - 1; buff < bp; bp-- ){
			ch = *bp;
			if( normalize && SPorHT(ch) ){
				sv1log("##HHu reduce trail spaces<<%s<<\n",hp);
				*bp = 0;
			}else
			if( !CRorLF(ch) ){
				bp++;
				break;
			}
		}
		rsize = size - (bp - buff); 
		Normalized++;
	}
	return rcode;
}
/*
 * this function will not detect nor fix multiple anomaly in a line ...
 */
static check_header(Conn,head)
	Connection *Conn;
	char *head;
{	char ch,*wp,*hp;
	char tmp[URLSZ];
	int nctl,nbin,leng;
	int bad = 0;

	nctl = 0;
	nbin = 0;
	for( hp = head; ch = *hp; hp++ ){
		if( ch & 0x80 )
			nbin++;
		else
		if( ch < 0x20 && ch != '\t' && ch != '\r' &&  ch != '\n' )
			nctl++;
	}
	leng = hp - head;
	if( nctl || nbin || 256 < leng ){
		strfConn(Conn,"%h",tmp);
		daemonlog("F",
		"Suspicious HTTP header [%s][CTL=%d,BIN=%d,LEN=%d] %s\n",
			tmp,nctl,nbin,hp-head,head);
	}

	for( hp = &head[leng]; head+1 < hp && CRorLF(hp[-1]); hp-- );
	if( head+1 < hp && CRorLF(*hp) ){
		wp = hp - 1;
		if( SPorHT(*wp) ){
			for( ; head+1 < wp && SPorHT(wp[-1]); wp-- );
			sv1log("##HHn remove space before CRLF<<%s",head);
			strcpy(wp,hp);
		}
	}

	ch = head[0];
	if( CRorLF(ch) ){
		/* End of header */
	}else
	if( SPorHT(ch) ){
		sv1log("##HHe bad space at the top of header<<%s",head);
		bad =
		BadRequest = 1;
	}else
	if( ch == ':' ){
		sv1log("##HHe empty field-name<<%s",head);
		bad =
		BadRequest = 2;
	}else{
	  for( hp = head; ch = *hp; hp++ ){
	    if( ch == ':' ){
		wp = ++hp;
		while( SPorHT(*hp) )
			hp++;
		if( hp == wp ){
			sv1log("##HHn insert space after ':'<<%s",head);
			strcpy(tmp,hp);
			sprintf(hp," %s",tmp);
			Normalized++;
		}else
		if( hp != wp + 1 ){
			sv1log("##HHn reduce space after ':'<<%s",head);
			strcpy(wp+1,hp);
			Normalized++;
		}
		break;
	    }
	    if( CRorLF(ch) ){
		sv1log("##HHe lacking ':field-value'<<%s",head);
		bad =
		BadRequest = 3;
		break;
	    }
	    if( SPorHT(ch) ){
		wp = hp;
		while( ch = *++hp )
			if( !SPorHT(ch) )
				break;
		if( ch == ':' ){
			sv1log("##HHn remove space before ':'<<%s",head);
			strcpy(wp,hp);
			Normalized++;
		}else{
			sv1log("##HHe multi-tokens as a field-name<<%s",head);
			bad =
			BadRequest = 4;
		}
		break;
	    }
	  }
	}

	if( bad && !HTTP_rejectBadHeader ){
		sv1log("##HHn remove bad-header<<%s",head);
		head[0] = 0;
	}
}

#include <ctype.h>
static escape_request(Conn,request,size)
	Connection *Conn;
	unsigned char *request;
{	unsigned char *hostport,*xp,*hp;
	int encode,hc;

	hostport = strstr(request,"://");
	if( hostport == NULL )
		return;

	hostport += 3; 
	encode = 0;
	xp = OREQ_MSG;
	for( hp = hostport; hc = *hp; hp++ ){
		if( hc == 0 || hc == ' ' || hc == '\r' || hc == '\n'
		 || hc == '/' || hc == '?' )
			break;
		if( isalnum(hc) || hc == '-' || hc == '.' || hc == ':'
		 || hc == '@' || hc == '%' || strchr("$_!~*'(),;&=+#",hc) ){
			*xp++ = hc;
		}else{
			encode = 1;
			sprintf(xp,"%%%02x",hc);
			xp += 3;
		}
	}
	if( encode == 0 )
		return;
	strcpy(xp,hp);
	strcpy(hostport,OREQ_MSG);
	sv1log("## pre-escape request with URL including unsafe char.\n");
}

static char *fgetsRequest(Conn,buff,size,fc,normalize)
	Connection *Conn;
	char *buff;
	FILE *fc;
{	char *rcode;
	int len;

	if( 0 < OREQ_LEN )
	if( 30 < IO_TIMEOUT )
	if( WillKeepAlive )
	if( strncasecmp(OREQ_MSG,"POST",4) == 0 )
	{
		/* Mozilla/3.? stops sending in the middle of a POST
		 * request header after the connection is broken which
		 * is notified as Proxy-Connection: keep-alive ...
		 */
		int timeout;
		timeout = 5 * 1000;
		if( DDI_PollIn(Conn,fc,timeout) <= 0 ){
			sv1log("#### fgetsRequest: TIMEOUT\n");
			return NULL;
		}
	}

	rcode = fgets822FromC(Conn,buff,size,fc,normalize);

	if( rcode != NULL && OREQ_LEN == 0 )
		escape_request(Conn,buff,size);

	if( rcode != NULL && strtailchr(buff) != '\n' ){
		char tmp[64];
		Strncpy(tmp,buff,sizeof(tmp));
		sv1log("##HHe bad header ending without LF (leng=%d) [%s]...\n",
			strlen(buff),tmp);
		*buff = 0;
		rcode = NULL;
	}
	if( rcode != NULL && normalize )
		check_header(Conn,buff);

	if( rcode != NULL ){
		if( OREQ_LEN == 0 )
		if( HTTP_isMethod(buff) )
			HTTP_repairRequest(buff);
		len = strlen(buff);
		if( OREQ_LEN + len < sizeof(OREQ_MSG) ){
			strncpy(OREQ_MSG+OREQ_LEN,buff,len+1);
			OREQ_LEN += len;
		}
	}
	return rcode;
}
static clientAskedKeepAlive(Conn,fname,value)
	Connection *Conn;
	char *fname,*value;
{
	Strncpy(ConnFname,fname,sizeof(ConnFname));
	if( strcasecmp(value,"keep-alive") == 0 ){
		ClntKeepAlive = 1; /* client requests so */
		if( !strcaseeq(REQ_METHOD,"POST")
		 && !DontKeepAlive && 1 < CKA_RemAlive )
			WillKeepAlive = 1; /* I will do so */
		else	WillKeepAlive = 0;
	}else
	if( strcasecmp(value,"close") == 0 ){
		clntClose(Conn,"c:client's will");
		ClntKeepAlive = 0;
		WillKeepAlive = 0;
	}
}
static char *addBuf(sp,xp,str)
	char *sp,*xp,*str;
{	int len;

	len = strlen(str);
	if( xp <= sp + len ){
		sv1log("##FATAL: addBuf over flow: %s\n",str);
		return sp;
	}
	strcpy(sp,str);
	return sp + len;
}
/*
 * This should be splited into relayRequestHead() and relayRequestBody()
 */
static relay_request(Conn,QX,request,fields,fc,ts,headonly,in_header,cpath,cdate)
	Connection *Conn;
	QueryContext *QX;
	char *request,*fields;
	FILE *fc,*ts;
	char *cpath;
{	char req[URLSZ];
	char value[256];
	int lines;
	int bytes;
	int sentFrom;
	int tsEOF;
	char *tail;
	char UA[1024],ViaBuf[2048],*vp,*vx,connection[256],fname[256],public[256];

	if( headonly && !in_header )
		return 0;

	sentFrom = 0;
	ClntIfMod[0] = 0;
	ClntIfModClock = -1;
	if( LockedByClient ){ free(LockedByClient); LockedByClient = 0; }
	tsEOF = 0;
	lines = bytes = 0;
	tail = fields;
	public[0] = 0;
	D_HOPS = 0;
	UA[0] = 0;
	ViaBuf[0] = 0;
	vp = ViaBuf;
	vx = ViaBuf + sizeof(ViaBuf);
	connection[0] = 0;

	if( in_header )
	for(;;){
		if( fgetsRequest(Conn,req,sizeof(req),fc,1) == NULL ){
			setClientEOF(Conn,fc,"relay_request");
			break;
		}
		lines += 1;
		bytes += strlen(req);

		value[0] = 0;
		if( STRH_Connection(req) ){
			scan_field1(req,fname,sizeof(fname),value,sizeof(value));
			clientAskedKeepAlive(Conn,fname,value);

			if( connection[0] != 0 )
				strcat(connection,",");
			strcat(connection,value);
			continue;
		}else
		if( strncasecmp(req,"Host:",5) == 0 ){
		}else
		if( strncasecmp(req,"Public-DeleGate:",16) == 0 ){
			lineScan(req+16,public);
		}else
		if( strncasecmp(req,"From:",5) == 0 ){
			sv1log("%s",req);
			if( strpbrk(req,"/(") )
				vp = addBuf(vp,vx,req);
		}else
		if( strncasecmp(req,"Forwarded:",10) == 0 ){
			vp = addBuf(vp,vx,req);
			D_HOPS++;
		}else
		if( strncasecmp(req,"Via:",4) == 0 ){
			vp = addBuf(vp,vx,req);
			D_HOPS++;
		}else
		if( strncasecmp(req,"User-Agent:",11) == 0 ){
			lineScan(req+11,UA);
		}else
		if( strncasecmp(req,"Referer:",8) == 0 ){
			/*sv1log("%s",req);*/
			/*stripDeleGate(Conn,req+8);*/
		}else
		if( strncasecmp(req,"Authorization:",14) == 0 ){
		}else
		if( strncasecmp(req,"Accept-Language:",16) == 0 ){
			HTTP_scanAcceptCharcode(Conn,req);
			if( *req == 0 )
				continue;
		}else
		if( HTTP_ignoreIf && strncasecmp(req,"If-",3) == 0 ){
			sv1log("IGNORE IF-: %s",req);
			continue;
		}else
		if( strncasecmp(req,"If-Modified-Since:",18) == 0 ){
			strcpy(ClntIfMod,req);
			lineScan(ClntIfMod+18,value);
			ClntIfModClock = scanHTTPtime(value);
			sv1log("= (%d) %s",ClntIfModClock,req);
			continue;
		}else
		if( strncasecmp(req,"Cookie:",7) == 0 ){
			char cookie[256];
			lineScan(req+7,cookie);
			if( strncmp(cookie,"DeleGate-Control=",17) == 0 ){
				sv1log("#Proxy-Cookie: %s",req);
				lineScan(cookie+17,proxyCookie);
			}
		}else
		if( strncasecmp(req,"Pragma:",7) == 0 ){
			Verbose("%s",req);
			wordScan(req+7,value);
			if( strcaseeq(value,"no-cache")){
				DontReadCache = 1;
				PragmaNoCache = 1;
				RES_CACHE_DISABLE = 1;
			}else
			if( strcaseeq(value,"cache-readonly") ){
				DontWriteCache = 1;
				continue;
			}else
			if( strcaseeq(value,"cache-only") ){
				CacheOnly = 1;
				DontWriteCache = 1;
				continue;
			}
		}else
		if( STRH(req,F_CacheControl) ){
		}else
		if( STRH(req,F_Range) ){
			sscanf(req,"%*s bytes=%d-%d",&QX_range[0],&QX_range[1]);
			sv1log("#HT11 (%d) %s",QX_range[1]-QX_range[0]+1,req);
			DontReadCache = 1;
			DontWriteCache = 1;
		}else
		if( strncmp(req,"X-Locking: ",11) == 0 ){
			lineScan(req+11,value);
			LockedByClient = strdup(value);
			continue;
		}

		if( *req == '\r' || *req == '\n' ){
			/* Append some fields at the end of header fields */
			if( !sentFrom ){
				sentFrom = 1;
/*
This should be the E-mail address of proxy for security consideration.
if( tail )tail = Sprintf(tail,"From: %s\r\n",D_FROM);
if( ts != NULL ) fprintf(ts,  "From: %s\r\n",D_FROM);
*/
			}
		}

if( tail )tail = Sprintf(tail,"%s",req);
if( ts != NULL ){
			if( fputs(req,ts) == EOF ){
				tsEOF = 1;
				break;
			}
		}

		if( *req == '\r' || *req == '\n' ){
			if( AcceptLanguages && AcceptLanguages[0] )
				Verbose("Accept-Language: %s\n",AcceptLanguages);
			Verbose("HTTP Relay_request_head (%d bytes/%d lines)\n",
				bytes,lines);
			break;
		}
	}

	if( in_header ){
		char *user,host[256],UaVia[4096],*dp;
		extern char *getClientUserC();
		int port,direct;

		strcpy(host,"-");
		port = getClientHostPort(Conn,host);

		for( dp = ViaBuf; *dp; dp++ ){
			switch( *dp ){
				case '\r': *dp = ';'; break;
				case '\n': *dp = ' '; break;
			}
		}
		direct = ViaBuf[0] == 0 && strstr(UA," via ") == NULL;
		sprintf(UaVia,"host=%s; User-Agent: %s; %s",host,UA,
			direct?"DIRECT":ViaBuf);
		if( !LOG_GENERIC || !fputLog(Conn,"Proxy","%s\n",UaVia) )
			    sv1log("Proxy: %s\n",UaVia);
		if( LOG_GENERIC ){
			if( (user = getClientUserC(Conn)) == NULL )
				user = "-";
			fputLog(Conn,"Client","%s@%s; agent=%s\n",user,host,UA);
		}

		if( connection[0] && RequestSerno == 0 )
		sv1log("HCKA:[%d] %s; host=%s; (%sUser-Agent: %s)\n",
			RequestSerno,connection,host,ViaBuf,UA);

		if( WillKeepAlive ){
			if( ImMaster ){
				if( vercmp(ClientVER,CHUNKED_VER) < 0 )
					clntClose(Conn,"v:via old DeleGate");
			}else
			if( !direct ){
				if( KeepAliveWithProxyClient ){
					sv1log("HCKA: Via=%s\n",ViaBuf);
				}else	clntClose(Conn,"v:via proxy");
			}
		}

		/* if this is the front most DeleGate for the client ... */
		if( !ImMaster )
			UAspecific(Conn,UA,direct);
	}

	if( public[0] ){
		keepPublic(Conn,public,fileno(fc));
		/* never return */
	}

	if( ts != NULL && !tsEOF ){
		if( !ClientEOF && !headonly ){
			lines = 0;
			DDI_proceedFromC(Conn,fc);
			bytes = relaydata(Conn,fc,ts,REQ_TIMEOUT);
		}
		fflush(ts);
	}

	Verbose("HTTP Relay_request done (%d bytes/%d lines)\n",bytes,lines);
	return lines;
}
/*
 * relayRequest in background is to simplify implementation of request
 * relay without knowing whether or not the request method has request body
 * which EOF could not be determined without knowledge, thus difficult to be
 * done in foreground before relaying the response.
 * HINT: HTTP header can be relayed in foreground without any knowledge :-D
 */
HTTP_methodWithBody(method)
	char *method;
{
	return !HTTP_methodWithoutBody(method);
}
static relayRequestBody1(Conn,fc,ts,fs,req,fields)
	Connection *Conn;
	FILE *fc,*ts,*fs;
	char *req,*fields;
{	char cleng[128];
	int bleng,bsize;
	int ch;
	int bn;
	char *bbuff;
	int nready;
	int start,lastflush,now;
	int buffered;

	if( getFV(fields,"Content-Length",cleng) == 0 )
		return 0;

	bleng = atoi(cleng);
	bsize = bleng + 1024;
	bbuff = malloc(bleng+1024);

	bn = DDI_flushCbuf(Conn,bbuff,bsize);
	DDI_proceedFromC(Conn,fc); /* it clears Cbuf ... */

	if( checkServ(Conn) ){
		buffered = 0;

		fputs(req,ts);
		fflush(ts);
		if( badServ(Conn,"beforeREQHEAD",fs,NULL) )
			ts = NULLFP();
		fputs(fields,ts);
		fflush(ts);
		if( badServ(Conn,"beforeREQBODY",fs,NULL) )
			ts = NULLFP();
	}else{
	buffered = 1;
	}
	start = lastflush = time(0);

	for( ; bn < bleng || 0 < READYCC(fc); bn++ ){
		if( bn < bleng && READYCC(fc) <= 0 ){
			nready = fPollIn(fc,(int)(HTTP_POLL_REQBODY*1000));
			if( nready <= 0 || HTTP_POLL_REQBODY < time(0) - lastflush ){
				now = time(0);
				if( buffered ){
					buffered = 0;
					fputs(req,ts);
					fputs(fields,ts);
					fwrite(bbuff,1,bn,ts);
					free(bbuff);
			sv1log("#### detach reqBuff: (%dB/%dB) / (%ds/%ds)\n",
					bn,bleng,now-lastflush,now-start);
				}
				fflush(ts);
				lastflush = now;
			}
			if( nready < 0 )
				setClientEOF(Conn,fc,"relayRequestBody-1");
			if( nready <= 0 ){
			sv1log("#### ERROR: IMMATURE REQUEST BODY (%d/%d)\n",
				bn, bleng);
				break;
			}
		}
		ch = getc(fc);
		if( ch == EOF ){
			setClientEOF(Conn,fc,"relayRequestBody-2");
			break;
		}
		if( buffered ){
			if( bsize <= bn ){
				bsize += 1024;
				bbuff = (char*)realloc(bbuff,bsize);
			}
			bbuff[bn] = ch;
		}else{
			putc(ch,ts);
		}
		if( OREQ_LEN < sizeof(OREQ_MSG) )
		OREQ_MSG[OREQ_LEN++] = ch;

		if( ch == '\n' && checkServ(Conn) ){
			fflush(ts);
			if( badServ(Conn,"inBODY",fs,NULL) )
				ts = NULLFP();
		}
	}
	if( OREQ_LEN < sizeof(OREQ_MSG) )
	OREQ_MSG[OREQ_LEN] = 0;

	if( buffered ){
		bbuff[bn] = 0;
		/* do necessary rewriting for the body, the header fields
		 * and request line here, or return body data to the caller
		 * of this function without sending to the server.
		 */
		fputs(req,ts);
		fputs(fields,ts);
		fwrite(bbuff,1,bn,ts);
		free(bbuff);
	}

	fflush(ts);
	sv1log("relayRequestBody1 done (%d/%d)\n",bn,bleng);
	return 1;
}
redirectXdisplay(Conn,fc,method,req,fields,bodyp,blengp)
	Connection *Conn;
	FILE *fc;
	char *method,*req,*fields,**bodyp;
	int *blengp;
{	int xpid;
	int bleng;
	int rcc;
	char *body;
	char cleng[256],uagent[256],ctype[256],display[256],xproxy[256],*rem;
	char edisplay[256],exproxy[256];
	char myname[256],peername[256];
	char me[256],myhp[256];

	if( !strcaseeq(method,"POST") )
		return 0;

	if( getFV(fields,"User-Agent",uagent) == NULL )
		return 0;
	if( strstr(uagent,"X11") == NULL && strstr(uagent,"X Window") == NULL)
		return 0;

	if( getFV(fields,"Content-Type",ctype) == NULL )
		return 0;
	if( strstr(ctype,"application/") != ctype )
		return 0;
	if( strstr(ctype,"www-form-urlencoded") == NULL )
		return 0;

	if( getFV(fields,"Content-Length",cleng) == NULL )
		return 0;
	bleng = atoi(cleng);
	if( bleng == 0 )
		return 0;

	body = malloc(bleng+100);
	DDI_proceedFromC(Conn,fc);
	rcc = fread(body,1,bleng,fc);
	body[rcc] = 0;
	*bodyp = body;
	*blengp = rcc;
	if( rcc != bleng ){
		sv1log("#### Content-Length: %d != real length=%d\n",bleng,rcc);
		bleng = rcc;
	}

	if( 0 < rcc ){
		strncpy(OREQ_MSG+OREQ_LEN,body,rcc+1);
		OREQ_LEN += rcc;
	}

	if( strncmp(body,"proxy_request=",14) == 0 ){
		char xbody[4096];
		nonxalpha_unescape(body,xbody,0);
		Verbose("#### POST REQUEST BODY ####\n%s\n",xbody);
		strcpy(req,"GET ");
		strcpy(req+4,req+5);
		Verbose("#### REQUEST ####\n%s",req);
		return 0;
	}

	if( strncmp(body,"display=",8) != 0 )
		return 0;

	paramscanX(body+8,"&",edisplay,sizeof(edisplay));
	nonxalpha_unescape(edisplay,display,0);
	if( strchr(display,':') == NULL )
		return 0;

	rem = strdup(body+8+strlen(edisplay));
	gethostNAME(ToS,myname);
	getpeerNAME(ToS,peername);
	ClientIF_HP(Conn,myhp);
	sprintf(me,"%s://%s/",DFLT_PROTO,myhp);
	xpid = makeXproxy(Conn,xproxy,display,myname,peername,me,60);
	sv1log("#### Xproxy[%d]: %s <- %s <- %s\n",xpid,display,xproxy,peername);

	if( xpid == 0 )
		return 0;

	nonxalpha_escape(xproxy,exproxy);
	sprintf(body,"display=%s%s",exproxy,rem);
	free(rem);
	bleng += strlen(exproxy) - strlen(edisplay);
	sprintf(cleng,"%d",bleng);
	replaceFieldValue(fields,"Content-Length",cleng);

	*bodyp = body;
	*blengp = bleng;
	return xpid;
}
static relayBuffered(in,out)
	FILE *in,*out;
{	int rcc;
	char buf[1024];

	fflush(out);
	while( rcc = fgetBuffered(buf,sizeof(buf),in) ){
		if( Fwrite(buf,1,rcc,out) <= 0 )
			return;
	}
	while( 0 < PollIn(fileno(in),1000) ){
		rcc = read(fileno(in),buf,sizeof(buf));
		if( rcc <= 0 )
			break;
		if( write(fileno(out),buf,rcc) != rcc )
			break;
	}
}
static genProxyReqFields(Conn,fields,cpath)
	Connection *Conn;
	char *fields;
	char *cpath;
{	char *gfp;
	char forwarded[256],received_by[256];
	char *auth,atype[128],genauth[256];
	char buf[1024],buf2[1024];
	char *dp;
	int withAuth;

	gfp = fields;
	*gfp = 0;

	if( findFieldValue(REQ_FIELDS,"From") == NULL )
	if( makeFrom(Conn,buf) ){
		Verbose("## GEN From: %s\n",buf);
		gfp = Sprintf(gfp,"From: %s\r\n",buf);
	}
	if( makeForwarded(Conn,forwarded) ){
		Verbose("## GEN Forwarded: %s\n",forwarded);
		gfp = Sprintf(gfp,"Forwarded: %s\r\n",forwarded);
	}
	HTTP_genVia(Conn,1,received_by);
	gfp = Sprintf(gfp,"Via: %s\r\n",received_by);

	withAuth = 0;
	if( auth = findFieldValue(REQ_FIELDS,"Authorization") )
	if( HTTP_decompAuth(auth,atype,sizeof(atype),buf,sizeof(buf)) ){
		if( buf[0] == 0 || buf[0] == ':' )
			sv1log("## ignore empty Authorization [%s]\n",buf);
		else{
			Verbose("## PASS Authorization: %s %s\n",atype,"");
			withAuth = 1;
		}
	}
	if( IsMounted && strchr(REAL_SITE,'@') ){
		strcpy(buf2,REAL_SITE);
		*strchr(buf2,'@') = 0;
		str_to64(buf2,strlen(buf2),buf,sizeof(buf),1);
		if( dp = strpbrk(buf,"\r\n") )
			*dp = 0;
		strcpy(atype,"Basic");
		sprintf(genauth,"%s %s",atype,buf);
		sv1log("## MOUNT Authorization: %s [%s]\n",genauth,buf2);
		gfp = Sprintf(gfp,"Authorization: %s\r\n",genauth);
	}else
	if( withAuth == 0 )
	if( makeAuthorization(Conn,genauth,0) ){
		HTTP_decompAuth(genauth,atype,sizeof(atype),buf2,sizeof(buf2));
		sv1log("## GEN Authorization: %s [%s]\n",genauth,buf2);
		gfp = Sprintf(gfp,"Authorization: %s\r\n",genauth);
	}
	if( makeAuthorization(Conn,genauth,1) ){
		HTTP_decompAuth(genauth,atype,sizeof(atype),buf2,sizeof(buf2));
		sv1log("## GEN Proxy-Authorization: %s [%s]\n",genauth,buf2);
		gfp = Sprintf(gfp,"Proxy-Authorization: %s\r\n",genauth);
	}

	if( cpath != NULL )
	if( toProxy || toMaster )
	{	char fileId[URLSZ];

		getFileUID(fileId,cpath);
		gfp = Sprintf(gfp,"X-Locking: %s\r\n",fileId);
	}
}
static codeconv_reqhead(Conn)
	Connection *Conn;
{	char *method,*ver,*sp;
	HttpRequest reqx;
	char url1[URLSZ],url2[URLSZ];
	int oc;

	if( CCXactive(CCX_TOSV) == 0 )
		return;

	decomp_http_request(REQ,&reqx);
	method = reqx.hq_method;
	strcpy(url1,reqx.hq_url);
	ver = reqx.hq_ver;
	

	if( sp = strchr(url1,'?') ){
		strcpy(url2,sp+1);
		sprintf(sp,"-_-_-%s",url2);
	}
	nonxalpha_unescape(url1,url2,0);
	oc = CCXexec(CCX_TOSV,url2,strlen(url2),url1,sizeof(url1));
	CCXexec(CCX_TOSV,"",0,url1+oc,sizeof(url1)-oc);

	if( strcmp(url1,url2) != 0 ){
		nonxalpha_escape(url1,url2);
		if( sp = strstr(url2,"-_-_-") )
			sprintf(sp,"?%s",sp+5);
		sprintf(url1,"%s %s HTTP/%s\r\n",method,url2,ver);
		sv1log("applied char. code conv. for request header (%d<-%d)\n",
			strlen(url1),strlen(REQ));
		strcpy(REQ,url1);
		Verbose(">> %s",REQ);
	}
}

static ReqVer(req1,v1,req2,v2)
	char *req1,*req2;
{	char *ver;

	if( ver = strstr(req1," HTTP/1.") )
	if( ver[8] == v1 )
	if( streq(ver+9,"\r\n") || streq(ver+9,"\n") )
	{
		ver += 8;
		if( req1 != req2 ){
			strcpy(req2,req1);
			ver = req2 + (ver - req1);
		}
		if( *ver != v2 ){
			*ver = v2;
			Verbose("#HT11 1.%c -> 1.%c: %s",v1,v2,req2);
		}
		if( streq(ver+1,"\n") ){
			strcpy(ver+1,"\r\n");
			sv1log("##HHn replaced LF to CRLF: %s",req2);
		}
		return 1;
	}
	return 0;
}

static relayRequest(Conn,QX,cpath,cdate,fc,ts,fs)
	Connection *Conn;
	QueryContext *QX;
	char *cpath;
	FILE *fc,*ts,*fs;
{	register int ppid,cpid;
	char method[128];
	int in_header;
	int with_reqbody;
	char *body;
	int bleng;
	int xpid;
	char genfields[0x8000];
	char req[URLSZ];

	if( badServ(Conn,"beforeREQ",fs,NULL) )
		ts = NULLFP();

	if( CCXactive(CCX_TOSV) )
		codeconv_reqhead(Conn);

	if( toProxy ){
		makeProxyRequest(Conn,REQ,REQ);
	}

	if( !HTTP_reqWithHeader(Conn,REQ) ){
		fputs(REQ,ts);
		if( !HTTP_isMethod(REQ) ){
			relayBuffered(fc,ts);
			/* should be relayed by parallel thread/process */
		}
		fflush(ts);
		return 0;
	}

	wordScan(REQ,method);
	with_reqbody = HTTP_methodWithBody(method);

	genProxyReqFields(Conn,genfields,cpath);
	MountReferer(Conn,REQ_FIELDS);
	strcat(genfields,REQ_FIELDS);

	if( with_reqbody ){
		if( relayRequestBody1(Conn,fc,ts,fs,REQ,genfields) )
			return 0;
		if( ClientEOF )
			return 0;

		ppid = getpid();
		if( cpid = Fork("relayRequest") ){
			if( cpid == -1 ){
				/* should send TOO BUSY message to the client */
				sv1tlog("CANNOT FORK\n");
				Finish(1);
			}
			return cpid;
		}
		setRelaying(NULL);
		Vsignal(SIGTERM,sigTERM);
	}

bleng = 0;
body = NULL;
xpid = redirectXdisplay(Conn,fc,method,REQ,REQ_FIELDS,&body,&bleng);

	if( toMaster && vercmp(MediatorVer,CHUNKED_VER) < 0 ){
		/* DeleGate/5.9.3 or older don't support HTTP/1.1 but
		 * don't rewrite HTTP/1.1 to HTTP/1.0 when acting as
		 * a MASTER DeleGate.
		 */
		if( ReqVer(REQ,'1',req,'0') ){
			fputs(req,ts);
			sv1log("#HT11 FORCE HTTP/1.0 for MASTER-DeleGate/5\n");
		}else	fputs(REQ,ts);
	}else
	if( HTTP11_toserver && ReqVer(REQ,'0',req,'1') ){
		fputs(req,ts);
		fputs("Connection: keep-alive\r\n",ts);
		sv1log("#HT11 FORCE HTTP/1.1 or Connection:keep-alive\n");
	}else{
		fputs(REQ,ts);
	}
	fputs(genfields,ts);

	if( body != NULL ){
		fwrite(body,1,bleng,ts);
		free(body);
	}
	Verbose("HTTP relayed request %dhead+%dbody\n",strlen(REQ_FIELDS),bleng);

	if( with_reqbody ){
		relay_request(Conn,QX,REQ,0,fc,ts,0,0,cpath,-1);
		if( feof(fc) ){
			sv1log("Kill (%d/%d, SIGTERM)\n",ppid,getppid());
			Kill(ppid,SIGTERM);
		}else{
			Verbose("C-S relay TIMEOUT\n");
		}
		/* if( xpid ) Kill(xpid,SIGTERM); */
		Finish(0);
	}else{
		if( ts != NULL )
			fflush(ts);
		return 0;
	}
}

static recvRequestFields(Conn,QX,fc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc;
{
	if( REQ_FIELDS[0] != 0 ){
		/* already got */
		return;
	}
	if( HTTP_reqIsHTTP(Conn,REQ) ){
		if( HTTP_reqWithHeader(Conn,REQ) )
			relay_request(Conn,QX,REQ,REQ_FIELDS,fc,NULL,1,1,NULL,-1);
		else	DontWriteCache = 1;
	}else{
		/* a Gopher request without such request fields */
	}
}

static decomp_gopherURL(Conn,req,rpath)
	Connection *Conn;
	char *req,*rpath;
{	int gtype;
	char *path;
	HttpRequest reqx;

	path = reqx.hq_url;
	if( decomp_http_request(req,&reqx) ){
	/* Gopher/HTTP where path is URL with gtype */
		/* from client expecting proxy */
		if( path[0] == '/' && path[1] != 0 ){
			/* this should be gtype in URL */
			gtype = path[1];
			strcpy(path,path+2);
		}else
		/* from client without proxy support */
	    		gtype = get_gtype(path,path);

		if( rpath ) strcpy(rpath,path);
		CTX_set_clientgtype(Conn,gtype);
		return gtype;
	}else{
	/* Gopher/Gopher */
		linescanX(req,path,sizeof(reqx.hq_url));
	    	gtype = get_gtype(path,path);
		if( rpath ) strcpy(rpath,path);
		CTX_set_clientgtype(Conn,gtype);
		return 0;
	}
}
static char *rewrite_request(Conn,fp)
	Connection *Conn;
	FILE *fp;
{	char *url,*urltop;
	char rproto[256],rhost[512];
	int riport;
	char *cproto;
	char *rcode = REQ;
	int dgurl;
	char *ver;

	sv1log("REQUEST %s %s",BORN_SPECIALIST?"-":"=",REQ);

strcpy(rproto,"http");
riport = 80;
rhost[0] = 0;

	if( BORN_SPECIALIST || ACT_SPECIALIST ){
		if( 0 < REQ_VNO ){
			url = REQ + strlen(REQ_METHOD);
			while( *url == ' ' || *url == '\t' )
				url++;
			cproto = "http";
		}else{
			url = REQ;
			cproto = "gopher";
		}
		urltop = url;

		/*
		 * This stuff should be passed multiple times
		 * for indirect MOUNT, but it should be executed
		 * only for the first original reuqest URL.
		 */
		if( CLIENTS_PROXY[0] == 0 ){
			if( fromProxyClient(url) ){
				setProxyOfClient(Conn,1,url);
				DONT_REWRITE = 1;
			}else	setProxyOfClient(Conn,0,NULL);
		}

		MountRequestURL(Conn,url);
		rhost[0] = 0;
		dgurl = 0;

/*
 * How to treat a request sent form a proxy client like "http://proxy/-_-URL"
 * should be cleary defined... (^_^;
 *
 * 951113: such URL (and optional flags) sould be passed through to the
 * target Proxy in the form  /-_-=flags=URL  because the flags should be
 * used in the Proxy (DeleGate).
 */

		if( strncasecmp(url,"pop://",6) == 0
		 && strchr(" \t\r\n",url[6]) ){
			strcpy(rproto,"pop");
			strcpy(rhost,"*");
			riport = 110;
			set_realserver(Conn,rproto,rhost,riport);
		}else
		if( strncasecmp(url,"news:",5) == 0
		 && strncmp(url+5,"//",2) != 0 ){
			strcpy(rproto,"nntp");
			strcpy(rhost,"*");
			riport = 119;
			set_realserver(Conn,rproto,rhost,riport);
		}else
		if( URL_toMyself(Conn,url) == NULL ){
			Verbose("To another server or proxy, THRU >>> %s",url);
		}else
		if( CTX_url_derefer(Conn,cproto,url,Modifier,DELEGATE_FLAGS,rproto,rhost,&riport) ){
			dgurl = 1;
			set_realserver(Conn,rproto,rhost,riport);
			if( !IsMounted )
			if( !do_RELAY(Conn,RELAY_DELEGATE) ){
				sv1log("Forbidden: RELAY DELEGATE\n");
				RelayForbidden |= RELAY_DELEGATE;
/*
 * can't strictly forbid the relaying if some host is allowed ...
 * (it's not problem if the restriction is all or nothing way)
 *
 * It's so expensive to check all of URLs in the RESPONSE about whether or
 * not they are RELAYable.  So all of them are rewritten if the URL in the
 * REQUEST is RELAYable, thus access to them will visit the delegated again.
 * And they should be relayed because they are directed to the delegated by
 * the delegated.
 * But RESPONSE to them will not rewritten if they are not RELAYable, thus
 * the recursive relay by URL rewriting of delegated will stop there.
 */
			}else	DO_DELEGATE = 1;
		}

		if( url_deproxy(Conn,REQ,url,rproto,rhost,&riport) ){
			set_realserver(Conn,rproto,rhost,riport);
			if( !IsMounted )
			if( !do_RELAY(Conn,RELAY_PROXY) ){
				RelayForbidden |= RELAY_PROXY;
				sv1log("Forbidden: RELAY PROXY\n");
			}
		}

		/*
		 * cancel `DONT_REWRITE' if the URL is in form http://myself/-_-url
		 */
		if( ToMyself && dgurl )
			DONT_REWRITE = 0;

		if( rhost[0] == 0 )
			ToMyself = 1;

		if( strchr(" \t\r\n",url[0]) ){
			Strins(url,"/");
			Verbose("HTTP EMPTY URL to %s",REQ);
		}

		if( REAL_HOST[0] )
			Verbose("REMOTE > %s",REQ);

		wordScan(urltop,REQ_URL);
	}

	if( rproto[0] == 0 ){
		sv1log("What? %s",REQ);
		return NULL;
	}

	if( !strcaseeq(rproto,"https") )
	if( !strcaseeq(rproto,"http") ){
		IAM_GATEWAY = 1;

		add_localheader(Conn,DONT_REWRITE);
		add_DGinputs(Conn,"%s",REQ);
		sv1log("HTTP GateWay > %s://%s:%d/%s",rproto,rhost,riport,REQ);

		if( protoGopher(rproto) )
			decomp_gopherURL(Conn,REQ,NULL);
	}

	return rcode;
}

/*
 *     MOUNT="URL1/* URL2/* indirect"
 *     URL1/name/path
 *  -> URL2/name/indirect.cfi [Base: URL3]
 *  -> URL3/path
 * thus, add
 *     MOUNT="URL1/name/* URL3/*"
 * redirection for response, and bypass mount for succeeding request.
 */

#define DELIM "?_?"

static indirect_mount(Conn,req,tc)
	Connection *Conn;
	char *req;
	FILE *tc;
{	FILE *mtab;
	char *tp,*bp,*method,*mnt_url,*ver;
	HttpRequest reqx;
	char hostport[256],murl[URLSZ],rel_url[URLSZ],*rurl;
	char line[URLSZ],Base[URLSZ];
	char ourl[URLSZ],vurl[URLSZ],durl[URLSZ],*up;
	int nocache;
	char *upath;

	tp = strstr(req,DELIM);
	if( tp == NULL )
		return 0;
	strcpy(rel_url,tp+strlen(DELIM));

	decomp_http_request(req,&reqx);
	method = reqx.hq_method;
	mnt_url = reqx.hq_url;
	ver = reqx.hq_ver;

	if( bp = strstr(mnt_url,DELIM) )
		*bp = 0;
	if( bp = strpbrk(rel_url," \t\r\n") )
		*bp = 0;

	mtab = TMPFILE("indirect_mount");
	HostPort(hostport,DST_PROTO,DST_HOST,DST_PORT);
	sprintf(murl,"%s://%s%s",DST_PROTO,hostport,mnt_url);
	nocache = PragmaNoCache;
	URLget(murl,nocache,mtab);

/*
	if( fgets(line,sizeof(line),mtab) == NULL
	 || strncasecmp(line,"#!cfi",5) != 0 ){
		fclose(mtab);
		return 0;
	}
*/

	Base[0] = 0;
	while( fgets(line,sizeof(line),mtab) != NULL ){
		if( strncasecmp(line,"BASE:",5) == 0 )
			wordScan(line+5,Base);
	}
	fclose(mtab);
	if( Base[0] == 0 )
		return 0;
	if( strtailchr(Base) != '/' )
		strcat(Base,"/");

	CTX_url_rurl(Conn,murl,ourl,"http","-.-",0,"",0);
	if( upath = URL_toMyself(Conn,ourl) ){
		strcpy(vurl,upath);
		if( up = strrchr(vurl,'/') )
			up[1] = 0;
		strcat(vurl,"*");
		sprintf(durl,"%s*",Base);
		set_MOUNT_ifndef(Conn,vurl,durl,"");
		init_mtab();
	}

	/*
	 * the following may also be done by rewriting ORIGINAL request
	 * with the above MOUNT.
	 */
	rurl = rel_url;
	if( rurl[0] == '/' )
		rurl++;
	sprintf(req,"%s %s%s HTTP/%s\r\n",method,Base,rurl,ver);
	sv1log("#### %s",req);

	return 1;
}

FILE *insertFTOCL_X(Conn,tc)
	Connection *Conn;
	FILE *tc;
{	int toC;
	FILE *ntc;

	toC = ToC;
	setConnX(Conn,FromC,ToC,FromS,ToS);
	if( ToC == toC ){
		/* filter is not inserted */
		return tc;
	}else{
		ntc = fdopen(ToC,"w");
		/* if( fileno(tc) != ClientSock )
		/* fclose(tc); */
		return ntc;
	}
}


#define FORWARD_AUTH "forward-auth"
static strmatch(s1,s2) char *s1,*s2; { return strcmp(s1,s2) == 0; }
static forwardOK(Conn,what,fname)
	Connection *Conn;
{	char forw[1024];

	if( 0 <= find_CMAP(Conn,what,forw) ){
		if( scan_commaList(forw,0,strmatch,fname) )
			return 1;
	}
	return 0;
}
/*
 * reply 302 Moved "/-_-proto://server/" to the "/-_-proto://server" request
 */
static delegate_moved(Conn,tc)
	Connection *Conn;
	FILE *tc;
{	char *dp;
	char url[URLSZ];
	int totalc;

	if( dp = strchr(REQ,' ') )
	if( dp[1] == '/' && strchr(" \t\r\n",dp[2]) )
	if( dp = strchr(OREQ_MSG,' ') ){
		wordScan(dp+1,url);
		if( strtailchr(url) != '/' ){
			strcat(url,"/");
			totalc = putMovedTo(Conn,tc,url);
			http_Log(Conn,302,CS_INTERNAL,REQ,totalc);
			return 1;
		}
	}
	return 0;
}

static CacheClose(QX)
	QueryContext *QX;
{
	if( QX_cachefp ){
		fflush(QX_cachefp);
		lock_unlock(fileno(QX_cachefp));
		stopDistribution(QX_Conn,QX_cachefp,QX_cpath);
		fclose(QX_cachefp);
		QX_cachefp = NULL;
	}
}
static decompREQUEST(Conn,QX)
	Connection *Conn;
	QueryContext *QX;
{
	if( REAL_HOST[0] ){
		QX_proto  = REAL_PROTO;
		QX_site = REAL_HOST;
		QX_port  = REAL_PORT;

		if( REAL_SITE[0] == 0 ){
			HostPort(QX_site_buf,REAL_PROTO,REAL_HOST,REAL_PORT);
			sv1log("#### REAL_SITE = empty => [%s]\n",QX_site_buf);
		}else	strcpy(QX_site_buf,REAL_SITE);
		QX_site = QX_site_buf;

		log_PATH(Conn,">");
	}else{
		if( ImMaster )
			log_PATH(Conn,">");
		QX_proto = "http";
	}

	ProcTitle(Conn,"%s://%s/)",QX_proto,QX_site);
	sv1log("REQUEST = %s[%s://%s/] %s",PragmaNoCache?"(no-cache)":"",
		QX_proto,QX_site,REQ);
}
static lookCacheOnly(Conn,tc,fc)
	Connection *Conn;
	FILE *tc,*fc;
{	char *proto,*host;
	int port;
	char *req,*method,*url,*ver;
	HttpRequest reqx;
	char uproto[64],usite[256],upath[1024],cpath[URLSZ];
	int useCache;
	FILE *cachefp;
	int expire,cdate;
	char smtime[URLSZ];
	int mtime;

	cachefp = NULL;

	if( DontReadCache ){
		returnAckCANTCON(Conn,tc,"don't read cache");
		return;
	}

	proto = DFLT_PROTO;
	host = DFLT_HOST;
	port = DFLT_PORT;
	req = D_REQUEST;
	if( req == NULL ){
		sv1log("lookCacheOnly: no D_REQUEST set.\n");
		return;
	}

	decomp_http_request(req,&reqx);
	method = reqx.hq_method;
	url = reqx.hq_url;
	ver = reqx.hq_ver;
	decomp_absurl(url,uproto,usite,upath,sizeof(upath));

	sv1log("#### [%s][%s][%d] [%s][%s][%s]\n",
		proto,host,port, method,url,ver);

	if( !service_permitted(Conn,proto) ){
		returnAckDENIED(Conn,tc,"not permitted");
		return;
	}

	useCache = CTX_cache_path(Conn,proto,host,port,upath,cpath);
	expire = 0x7FFFFFFF;

	if( useCache )
	if( cachefp = cache_fopen_rd("HTTP",cpath,expire,&cdate) )
	if( lock_sharedNB(fileno(cachefp)) == 0 )
	{
		if( CacheLastMod != 0 ){
			mtime = HTTP_getLastMod(smtime,sizeof(smtime),cachefp,cpath);
			sv1log("LastModified: %d <= %d ?\n",
				CacheLastMod,mtime);
			if( mtime <= CacheLastMod ){
				returnAckCANTCON(Conn,tc,"cache is obsolete");
				goto EXIT;
			}
		}
		returnAckOK(Conn,tc,"found in cache");
		relay_response_fromCache(Conn,proto,host,port,req,cpath,cachefp,tc,fc);
		goto EXIT;
	}
	returnAckCANTCON(Conn,tc,"not found in cache");
EXIT:
	if( cachefp != NULL )
		fclose(cachefp);
}

static checkAUTH(Conn,QX,fc,tc,proto,site,port,mandatory)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
	char *proto;
	char *site;
{	int auth,pauth;
	int totalc;

	if( !Conn->from_myself && NotREACHABLE(Conn,proto,site,port) ){
		service_permitted2(Conn,proto,0); /* cause delayReject */
		totalc = putHttpRejectmsg(Conn,tc,proto,site,port,REQ);
		http_Log(Conn,403,CS_AUTHERR,REQ,totalc);
		return -1;
	}

	 auth = auth_proxy_auth() || auth_origin_auth(Conn);
	pauth = auth_proxy_pauth();
	if( mandatory && (auth || pauth) ){
		if(  auth && !findFieldValue(REQ_FIELDS,"Authorization")
		 || pauth && !findFieldValue(REQ_FIELDS,"Proxy-Authorization")
		)	goto FAILED;
	}

	if( service_permitted2(Conn,proto,1) )
		return 0;

	if( auth || pauth ){
		if(HTTP_proxyAuthorized(Conn,REQ,REQ_FIELDS,pauth,tc)){
			char *fname;
			if( pauth )
				fname = "Proxy-Authorization";
			else	fname = "Authorization";
			if( !forwardOK(Conn,FORWARD_AUTH,fname) )
				HTTP_delRequestField(Conn,fname);
			return 0;
		}
	}

FAILED:
	service_permitted2(Conn,proto,0); /* cause delayReject */
	if( auth || pauth ){
		totalc = putNotAuthorized(Conn,tc,REQ,pauth,"</>","");
		http_Log(Conn,401,CS_AUTHERR,REQ,totalc);
	}else{
		totalc = putHttpRejectmsg(Conn,tc,proto,site,port,REQ);
		http_Log(Conn,403,CS_AUTHERR,REQ,totalc);
	}
	return -1;
}

static doCONNECT(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	char *rproto;
	char rhost[256];
	int rport;
	char ctype[128];
	FILE *ts;
	int vno;

	if( strncasecmp(REQ,"CONNECT",7) != 0 )
		return 0;

	rproto = "https";

	DDI_proceedFromC(Conn,fc);

	rport = scan_hostportX("",REQ+8,rhost,sizeof(rhost));
	if( rport <= 0 ){
		sv1log("CONNECT: Unknown Request? %s",REQ);
		return -1;
	}
	set_realserver(Conn,rproto,rhost,rport);
	if( checkAUTH(Conn,QX,fc,tc,rproto,rhost,rport,0) != 0 )
		return -1;

	Conn->from_myself = 1; /* permission already checked by checkAUTH() */
	set_realserver(Conn,"https",rhost,rport);
	setConnStart(Conn);

	if( connect_to_serv(Conn,FromC,ToC,0) < 0 ){
		putHttpCantConnmsg(Conn,tc,rproto,rhost,rport,OREQ);
		http_Log(Conn,500,CS_CONNERR,REQ,0);
		return -1;
	}

	setConnDone(Conn);

	if( toMaster || toProxy ){
		if( ToServ )
			ts = ToServ;
		else	ts = fdopen(ToS,"w");
		fputs(REQ,ts);
		fputs(REQ_FIELDS,ts);
		fflush(ts);
	}else{
		vno = HTTP_reqIsHTTP(Conn,REQ);
		if( 100 <= vno ){
			char line[256];
			fprintf(tc,"HTTP/%s 200 Connection established.\r\n",
				MY_HTTPVER);
			clntClose(Conn,"C:CONNECT method");
			if( getKeepAlive(Conn,line) )
				fprintf(tc,"%s\r\n",line);
			fprintf(tc,"\r\n");
			fflush(tc);
		}
	}

	ProcTitle(Conn,"https://%s:%d/)",rhost,rport);
	ctype[0] = 0;
	QX_totalc = relay_svcl(Conn,FromC,ToC,FromS,ToS);
	httpStat = CS_WITHOUTC;
	http_log(Conn,rproto,rhost,rport,REQ,200,ctype,QX_totalc,0,
		CONN_DONE-CONN_START,Time()-CONN_DONE);
	close(ToS);
	close(FromS);
	return 1;
}
static nonHTTP(Conn,ftc,request)
	Connection *Conn;
	FILE *ftc[2];
	char *request;
{	char method[32],methods[256],*dp,dc;
	FILE *fc,*tc;

	fc = ftc[0];
	tc = ftc[1];

	dp = wordScan(request,method);
	dc = *dp;
	if( dc != 0 && dc != ' ' && dc != '\t' && dc != '\r' && dc != '\n' )
		goto non_http;

	if( strcaseeq(method,"X-CACHE-GET") ){
		GET_CACHE = 1;
		strcpy(request,request+8);
		return 0;
	}
	if( HTTP_allowMethod1(Conn,request) )
		return 0;

	if( strcaseeq(method,"!SET") ){
		DELEGATE_setenv(fc,tc,request);
		return 1;
	}

	if( isHelloRequest(request)
	 || VSAP_isMethod(request)
	 || streq(method,"HELO")
	 || streq(method,"CPORT")
	 || streq(method,"PARAM")
	 || streq(method,"FTPGET")
	){
		sv1log("---- WARNING! HTTP switched to GENERALIST[%s]\n",
			method);
		signal(SIGPIPE,SIG_DFL);
		DDI_proceedFromC(Conn,fc);
		beGeneralist(Conn,fc,tc,request);
		ftc[0] = ftc[1] = 0; /* fclose() in beGeneralist() */
		return 1;
	}

non_http:
	sv1log("#### Method Not Allowed: %s",request);
	fprintf(tc,"HTTP/%s 405 Method Not Allowed\r\n",MY_HTTPVER);
	HTTP_allowMethods(Conn,methods);
	fprintf(tc,"Allow: %s\r\n",methods);
	fprintf(tc,"Connection: close\r\n");
	fprintf(tc,"\r\n");
	fprintf(tc,"Method Not Allowed: %s\r\n",request);
	return 1;
}

static gotNullRequest(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	char clhost[256];

	if( getClientHostPort(Conn,clhost) == 0 )
		strcpy(clhost,"?");
	sv1log("HTTP empty_request ? from %s (%d)\n",clhost,Conn->cl_count);
	http_Log(Conn,500,CS_ERROR,"ERR /empty_request",0);
	delayConnError(Conn,REQ);
}
static rewriteRequest(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	int nredirect = 0;

REDIRECT:
	if( rewrite_request(Conn,fc) == NULL ){
		gotNullRequest(Conn,QX,fc,tc);
		return -1;
	}
	if( indirect_mount(Conn,REQ,tc) ){
		if( 5 < ++nredirect ){
			sv1log("ERROR: REDIRECT LOOP ?\n");
			return -1;
		}
		/* variables which should be decided by the rewritten request..*/
		IAM_GATEWAY = 0;

		/* some variable including DONT_REWRITE
		 * should be decided by the original request.
		 */

		goto REDIRECT;
	}
	return 0;
}
static checkCookie(Conn)
	Connection *Conn;
{	char *val;

	if( val = findFieldValue(REQ_FIELDS,"Cookie") ){
		MountCookieRequest(Conn,OREQ,val);
		DontReadCache = 1;
		DontWriteCache = 1;
		withCookie = 1;
	}else	withCookie = 0;
	return 0;
}
static int DELAY_RESPCLOSE = 0;
static doMyself(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	int stcode;
	int mtime;
	char ctype[256];
	double start;

	if( isMYSELF(QX_site) )
		ToMyself = 1;

	if( strstr(REQ,"?-_-") )
		ToMyself = 1;

	if( !ToMyself )
		return 0;

	if( auth_origin_auth(Conn) ){
		if( checkAUTH(Conn,QX,fc,tc,QX_proto,QX_site,QX_port,1) != 0 )
			return 1;
	}

	/* is public thus permission is not necessary? */
	stcode = 200;
	start = Time();
	mtime = 0;
	if( QX_totalc = HttpToMyself(Conn,REQ,REQ_FIELDS,fc,tc,&stcode) ){
		CONN_DONE = CONN_START = start;
		if( httpStat == 0 )
			httpStat = CS_INTERNAL;
		ctype[0] = 0;
		http_log(Conn,DST_PROTO,DST_HOST,DST_PORT,REQ,
			stcode,ctype,QX_totalc,mtime, 0.0,Time()-start);
		if( DELAY_RESPCLOSE ){
			fflush(tc);
			sv1log("## SLEEP BEFORE CLOSE: %d\n",DELAY_RESPCLOSE);
			sleep(DELAY_RESPCLOSE);
		}
		return 1;
	}

	return 0;
}
static HttpFtp(Conn,QX,fc,tc,svsock,server,iport,req)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
	char *server,*req;
{	char *method,*path,*ver;
	HttpRequest reqx;
	char gpath[1024],gtype;
	char user[256],pass[256],auth[512];
	AuthInfo ident;

	Conn->body_only = !HTTP_reqWithHeader(Conn,req);
	decomp_http_request(req,&reqx);
	method = reqx.hq_method;
	path = reqx.hq_url;
	ver = reqx.hq_ver;

	lineScan(path,gpath);
	gtype = get_gtype(gpath,path);

	scan_url_userpass(server,user,pass,"anonymous");

	auth[0] = 0;
	if( HTTP_getAuthorization(Conn,0,&ident,0)
	 || HTTP_getAuthorization(Conn,1,&ident,0) )
		sprintf(auth,"%s:%s",ident.i_user,ident.i_pass);

	return httpftp(Conn,fc,tc,ver,method,svsock,
		auth,user,pass,server,iport,gtype,path);
}
static HttpFinger(Conn,vno,sv,server,iport,path)
	Connection *Conn;
{
	if( sv == -1 )
		return 0;
	return httpfinger(Conn,sv,server,iport,path,vno);
}
static http_gateway(Conn,QX,fc,tc,sv)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	char *proto = QX_proto;
	char *server = QX_site;
	int iport = QX_port;
	char *req = REQ;
	char *method,*path,*ver;
	HttpRequest reqx;
	char user[256],pass[256],request[URLSZ],rserver[256];
	AuthInfo ident;
	int vno;
	int msock;
	int rcode,rtotal;
	char ctype[128];
	char gtype;
	char *dp;

	Where = "Gatewaying";
	nsigPIPE = 0;

	vno = decomp_http_request(req,&reqx);
	method = reqx.hq_method;
	path = reqx.hq_url;
	ver = reqx.hq_ver;

	rcode = 200;
	ctype[0] = 0;
	rtotal = 0;
	setConnStart(Conn);
	msock = -1;
	DDI_proceedFromC(Conn,fc);

	/* if( strcaseeq(proto,"ftp") || strcaseeq(proto,"file") ) */
	if( strcaseeq(proto,"ftp") ){
		setConnDone(Conn);
		get_gtype(path,path);
		rtotal = HttpFtp(Conn,QX,fc,tc,sv,server,iport,req);
		sprintf(request,"%s %s://%s:%d%s%s",method,proto,server,iport,
			(*path=='/'?"":"/"),path);
		goto EXIT;
	}

	sprintf(request,"GET %s://%s:%d",proto,server,iport);
	if( *path == '/' )
		strcat(request,path);
	sprintf(request+strlen(request)," HTTP/%s",ver);

	if( strcaseeq(proto,"wais") ){
		setConnDone(Conn);
		rtotal = HttpWais(Conn,vno,sv,server,iport,path);
		/*goto EXIT;*//* log was done in HttpWais */
		return;
	}
	if( strcaseeq(proto,"finger") ){
		setConnDone(Conn);
		rtotal = HttpFinger(Conn,vno,sv,server,iport,path);
		goto EXIT;
	}
	if( strcaseeq(proto,"news") || strcaseeq(proto,"nntp")
	 || strcaseeq(proto,"pop") ){
		setConnDone(Conn);
		if( path[0] == '/' )
			strcpy(path,path+1);
		else
		if( strncasecmp(path,"news:",5) == 0 )
			strcpy(path,path+5);

		HTTP_getAuthorization(Conn,0,&ident,0);
		rtotal = HttpNews(Conn,vno,sv,server,iport,path,request,
				ident.i_user,ident.i_pass,
				ClntKeepAlive&&WillKeepAlive);

		if( dp = strchr(ident.i_user,'@') ){
			server = dp + 1;
		}
		goto EXIT;
	}
	if( strcaseeq(proto,"whois") ){
		if( 100 <= vno ){
			putHEAD(Conn,tc,200,"Whois/HTTP gateway",NULL,
				"text/plain",NULL,0,-1,-1);
			fputs("\r\n",tc);
			fflush(tc);
		}
		sprintf(req,"whois://%s:%d?%s\r\n",DST_HOST,DST_PORT,
			*path=='/'?path+1:path);
		DDI_pushCbuf(Conn,req,strlen(req));
		service_whois(Conn);
		goto EXIT;
	}

	if( protoGopher(proto) ){
		if( gtype = decomp_gopherURL(Conn,req,path) ){
			vno = HTTP_reqIsHTTP(Conn,req);
			rtotal = HttpGopher(Conn,vno,sv,server,iport,gtype,path);
			sprintf(request,"GET %s://%s:%d/%s",
				proto,server,iport,path);
			goto EXIT;
		}
	}
	sprintf(request,"GET %s://%s:%d/%s",proto,server,iport,path);

	if( 0 <= (msock = openMaster(Conn,sv,server,1)) ){
		setConnDone(Conn);
		if( protoGopher(proto) )
			rtotal = relay_svcl(Conn,-1,ToC,msock,-1,1,512);
		else	rtotal = relay_svcl(Conn,FromC,ToC,msock,msock,1,512);
	}else{
		sv1log("ERROR: cannot connect to a MASTER\n");
		rcode = 403;
		rtotal = 0;
	}

EXIT:
	http_log(Conn,proto,server,iport,request,rcode,ctype,rtotal,0,
		CONN_DONE-CONN_START,Time()-CONN_DONE);

	if( nsigPIPE ){
		clntClose(Conn,"p:premature client EOF on gatewaying");
		setClientEOF(Conn,tc,"gatewaying_%s.SIGPIPE",proto);
	}
}
static doFtpCached(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	char *method,*path,*ver;
	HttpRequest reqx;
	char user[256],host[256];
	int vno;

	vno = decomp_http_request(REQ,&reqx);
	method = reqx.hq_method;
	path = reqx.hq_url;
	ver = reqx.hq_ver;

	if( vno < 100 )
		return 0;
	if( !streq(method,"GET") )
		return 0;

	if( strchr(QX_site,'@') ){
		char *dp;
		user[0] = host[0] = 0;
		dp = paramscanX(QX_site,"@",user,sizeof(user));
		if( *dp == '@' )
			dp = wordscanX(dp+1,host,sizeof(host));
	}else{
		strcpy(user,"anonymous");
		strcpy(host,QX_site);
	}
	if( QX_totalc = httpftp_cached(Conn,tc,user,host,QX_port,path) ){
		http_Log(Conn,200,CS_HITCACHE,REQ,QX_totalc);
		return 1;
	}
	return 0;
}
static doCache(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	int cretry;
	int dontWaitCache;

	if( QX_useCache == 0 )
		return 0;

	cretry = 0;
	set_DG_EXPIRE(Conn,QX_expire);
	dontWaitCache = DontWaitCache;

retry_read:
	if( QX_cachefp != NULL )
		CacheClose(QX);
	QX_cdate = -1;
	if( 5 < cretry++ )
		return 0;

	if( !DontReadCache )
	if( QX_cachefp = cache_fopen_rd("HTTP",QX_cpath,QX_expire,&QX_cdate) ){
		if( LockedByC(Conn,QX_cpath,fc) ){ CacheClose(QX); return 0; }
		if( lock_for_rd("HTTP",cretry,QX_cpath,QX_cachefp) != 0 ){
			sleep(CACHE_RDRETRY_INTERVAL);
			goto retry_read;
		}
		if( HTTP_selfExpired(QX_cachefp) && !Conn->co_nonet ){
			CacheClose(QX);
		}else{
			setConnStart(Conn);
			CONN_DONE = CONN_START - 1;
			/* minus value represent no connection ... */

			if( relay_response_fromCache(Conn,
			    QX_proto,QX_site,QX_port,REQ,
			    QX_cpath,QX_cachefp,tc,fc) != 0 )
			{
				/* sent something wrong to client X-< */
			}
			return 1;
		}
	}

	if( !DontWriteCache )
	if( QX_cachefp = cache_fopen_rw("HTTP",QX_cpath) ){
		if( LockedByC(Conn,QX_cpath,fc) ){ CacheClose(QX); return 0; }
		if( file_lock_wr("HTTP",QX_cachefp) != 0 ){
			FILE *rfp;
			int updated;

			rfp = recvDistribution(Conn,QX_cpath,&updated);
			if( updated )
				goto retry_read;

			if( rfp != NULL ){
				setConnStart(Conn);
				CONN_DONE = CONN_START;
				relay_response(Conn,0,QX_proto,QX_site,QX_port,
					REQ,NULL,0,rfp,tc,fc,NULL,0);
				fclose(rfp);
				return 1;
			}
			if( !DontReadCache && !dontWaitCache ){
				sleep(CACHE_WRRETRY_INTERVAL);
				goto retry_read;
			}
			CacheClose(QX);
		}else{
			makeDistribution(Conn,QX_cachefp,QX_cpath);
			DontWaitCache = 1;
		}
	}else{
		/* readable but not writable */
		if( QX_cdate != -1 )
			QX_cdate = -1;
	}
	return 0;
}
static doMaxHops(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{
	if( HTTP_MAXHOPS < D_HOPS ){
		sv1log("#### MAX_HOPS %d < %d HOPS\n",HTTP_MAXHOPS,D_HOPS);
		fprintf(tc,"HTTP/%s 502 too many hops\r\n",MY_HTTPVER);
		fprintf(tc,"\r\n");
		fprintf(tc,"Too many hops: %d\r\n",D_HOPS);
		return 1;
	}
	return 0;
}
static pollRequestBody(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	char cleng[128];
	int nready;
	int timeout;

	if( getFV(REQ_FIELDS,"Content-Length",cleng) )
	if( 0 < atoi(cleng) )
	if( DDI_PollIn(Conn,fc,100) == 0 ){
		timeout = (int)(HTTP_WAIT_REQBODY * 1000);
		sv1log("#### start polling body of %s %dms\n",REQ_METHOD,
			timeout);
		nready = DDI_PollIn(Conn,fc,timeout);
		sv1log("#### polling done: %d\n",nready);
	}
}

extern int USE_HTTPMAIL;
static doHTTPMAIL(Conn,QX,fc,tc,expire)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	FILE *cfp;
	char url[URLSZ],*dp;
	int fromcache;

	if( !USE_HTTPMAIL )
		return 0;

	sprintf(url,"%s://%s:%d/%s",QX_proto,QX_site,QX_port,QX_upath);
	if( dp = strpbrk(url," \t\r\n\f") )
		*dp = 0;
	sv1log("TRY HTTPMAIL: %s\n",QX_upath);

	/* host may be the same with client ... (maybe only for testing...) */
	if( QX_cachefp ) lock_unlock(fileno(QX_cachefp));

	if( PragmaNoCache )
		expire = 0;
	cfp = hmget1(url,expire);

	if( cfp ){
		relay_response(Conn,0,QX_proto,QX_site,QX_port,REQ,
			QX_cpath,0,cfp,tc,NULL,QX_cachefp,QX_cdate!=-1);
		fclose(cfp);
		return 1;
	}
	return 0;
}
static tryICP(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	char msg1[64],msg2[64];

	if( select_icpconf(Conn,NULL) )
		connect_to_serv(Conn,FromC,ToC,0);
	if( FromS < 0 )
		return 0;

	sprintf(msg1,"ConnType=%c%s%s%s%s",
		ConnType?ConnType:'?',
		IAM_GATEWAY?"/GateWay":"",
		toProxy?"/Proxy":"",
		toMaster?"/Master":"",
		0<=Conn->ca_objsize?"/HIT_OBJ":"");

	if( ConnType == 'i' ){
		if( 0 <= Conn->ca_objsize ){
			sprintf(msg2,"RECEIVED HIT_OBJ=%d",
				Conn->ca_objsize);
		}else
		if( toProxy ){
			ConnType = 'p';
			strcpy(msg2,"FORWARDING to HTTP proxy");
			IAM_GATEWAY = 0;
			connected_to_proxy(Conn,REQ,FromS);
		}else
		if( toMaster ){
			ConnType = 'm';
			strcpy(msg2,"FORWARDING to MASTER DeleGate");
		}else{
			strcpy(msg2,"FORWARDING to ORIGIN server");
		}
	}else{
		msg2[0] = 0;
		if( toProxy ){
			ConnType = 'p';
			strcpy(msg2,"FORWARDING to HTTP proxy");
			IAM_GATEWAY = 0;
			connected_to_proxy(Conn,REQ,FromS);
		}
	}
	sv1log("(ICP) %s{%s}(%s:%d) %s",msg1,msg2,QX_site,QX_port,REQ);

	if( IAM_GATEWAY ){
		http_gateway(Conn,QX,fc,tc,FromS);
		return -1;
	}
	return 0;
}
static OpenProxy(Conn,proto,server,iport)
	Connection *Conn;
	char *proto,*server;
{
	if( !with_PROXY() && !withTmpProxy() )
		return -1;

	if( REAL_PROTO[0] )
		proto = REAL_PROTO;
	else	proto = DFLT_PROTO;

	if( protoGopher(proto) )
		if( get_gtype(REQ,0) == '7' )
			IsGopherSearch = 1;

	connect_to_proxy(Conn,REQ,proto,server,iport);

	if( FromS < 0 )
		IsGopherSearch = 0;

	return FromS;
}
static tryConnects(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{	int force_on_proxy;

	if( tryICP(Conn,QX,fc,tc) < 0 )
		return -1;

	if( FromS < 0 )
	if( Forwardit(Conn,FromC,ToC,0) ){
		if( FromS < 0 )
		{
			/* connection failed with forward server */
			return -1;
		}
		if( streq(GatewayProto,"http") ){
			IAM_GATEWAY = 0;
			ConnType = 'p';
			connected_to_proxy(Conn,REQ,FromS);
			sv1log("FORWARDING to HTTP proxy.\n");
		}else{
			/* GatewayProto == "delegate" */
			if( IAM_GATEWAY ){
				http_gateway(Conn,QX,fc,tc,FromS);
				return -1;
			}
		}
		if( GatewayPath && GatewayPath[0] ){
			/*
			replace_url(REQ,GatewayPath);
			must set reverse rewriting in the response...
			*/
		}
	}

	force_on_proxy = with_PROXY() && force_forward(Conn,QX_proto)
		 || withTmpProxy();
	if( FromS < 0 && force_on_proxy )
		OpenProxy(Conn,QX_proto,QX_site,QX_port);

	if( !streq(QX_proto,"nntp") && !streq(QX_proto,"news")
	 && !streq(QX_proto,"pop")  && !streq(QX_proto,"ftp") )
	if( FromS < 0 ){
		connect_to_serv(Conn,FromC,ToC,0);
		if( 0 <= FromS && IAM_GATEWAY ){
			http_gateway(Conn,QX,fc,tc,FromS);
			return -1;
		}
	}

	if( FromS < 0 && !force_on_proxy )
		OpenProxy(Conn,QX_proto,QX_site,QX_port);

	if( FromS < 0 && IAM_GATEWAY ){
		http_gateway(Conn,QX,fc,tc,-1);
		return -1;
	}
	if( FromS < 0 ){
		sv1log("cannot connect to server: %s://%s:%d\n",
			DST_PROTO,DST_HOST,DST_PORT);

		if( checkClientEOF(Conn,tc,"cant_connect") ){
			http_Log(Conn,500,CS_CONNERR,REQ,0);
		}else
		if( QX_cachefp != NULL && QX_cdate != -1 ){
			setConnDone(Conn);
			sv1log("makeshift with old cache (A).\n");
			httpStat = CS_MAKESHIFT;
			relay_response(Conn,0,QX_proto,QX_site,QX_port,
				REQ,QX_cpath,1,QX_cachefp,tc,NULL,NULL,1);
		}else{
			QX_totalc = putHttpCantConnmsg(Conn,tc,
				QX_proto,QX_site,QX_port,OREQ);

			/* when Pragma: no-cache */
			if( QX_cachefp != NULL && 0 < file_size(fileno(QX_cachefp)) ){
				int ooff = ftell(QX_cachefp);
				fseek(QX_cachefp,0,2);
				sv1log("preserved old cache (%d->%d) %s\n",
					ooff,ftell(QX_cachefp),QX_cpath);
			}
			http_Log(Conn,500,CS_CONNERR,REQ,0);
			HTTP_delayCantConn(Conn,REQ,"500 cannot connect\r\n",1);
		}
		if( QX_cachefp != NULL && file_size(fileno(QX_cachefp)) <= 0 )
			QX_cacheRemove = 1;
		return -1;
	}

	if( toMaster ){
		sv1log("HTTP -> (%s:%d) %s",QX_site,QX_port,REQ);
	}else{
		std_setsockopt(ToS);
		sv1log("HTTP => (%s:%d) %s",QX_site,QX_port,REQ);
	}

	if( checkClientEOF(Conn,tc,"connected") ){
		http_Log(Conn,500,CS_EOF,REQ,0);
		return -1;
	}

	/* do getsockopt(client) immediately after EOF check */
	expsockbuf(ToC,0,MAX_BUFF_SOCKSEND);

	set_keepalive(FromS,1);
	expsockbuf(FromS,MAX_BUFF_SOCKRECV,0);
	setConnDone(Conn);
	return 0;
}
static doTruncateCache(Conn,QX,fc,tc)
	Connection *Conn;
	QueryContext *QX;
	FILE *fc,*tc;
{
	if( QX_rcode == R_EMPTY_RESPONSE && QX_cdate != -1 ){
		sv1log("makeshift with old cache (B).\n ");
		Verbose("cache-file: %s\n",QX_cpath);

		lock_sharedNB(fileno(QX_cachefp));
		relay_response(Conn,0,QX_proto,QX_site,QX_port,
			REQ,QX_cpath,1,QX_cachefp,tc,fc,NULL,QX_cdate != -1);
	}else
	if( DontTruncate ){
		Verbose("DontTruncate: rcode=%d ftell:%d\n",
			QX_rcode,ftell(QX_cachefp));
	}else
	if( QX_rcode < 0 ){
		sv1log("rcode=%d unlink %s\n",QX_rcode,QX_cpath);
		QX_cacheRemove = 1;
	}else{
		if( httpStat == CS_STABLE ){
		}else{
			sv1log("%d bytes written to [%s]\n",
				ftell(QX_cachefp),QX_cpath);
			Ftruncate(QX_cachefp,0,1);
		}
	}
}
static closeServ(Conn,QX,ts,fs)
	Connection *Conn;
	QueryContext *QX;
	FILE *fs,*ts;
{
	delServ(Conn,fileno(ts),fileno(fs));
	fclose(ts);
	fclose(fs);
	FromS = -1;
	if( 0 < QX_cpid ){
		Kill(QX_cpid,SIGTERM);
		wait(0);
		QX_cpid = 0;
	}
	Conn->sv_reusing = 0;
}

static service_http2(Conn,ftc)
	Connection *Conn;
	FILE *ftc[2];
{	QueryContext QXbuf,*QX = &QXbuf;
	FILE *ts,*fs;
	FILE *respFile,*tc_sav;
	int chunksav;
	FILE *fc,*tc;

	fc = ftc[0];
	tc = ftc[1];

	httpStat = 0;
	httpStatX = 0;

	QX_Conn = Conn;
	QX_site = DFLT_HOST;
	QX_port = DFLT_PORT;

	QX_cpath[0] = 0;
	QX_cachefp = NULL;
	QX_cdate = -1;
	QX_cacheRemove = 0;

	QX_range[0] = QX_range[1] = -1;

	PragmaNoCache = 0;
	DontTruncate = 0;
	ToMyself = 0;

	CheckServ = 0;
	WaitServ = 0;
	BadServ = 0;

	if( ImMaster && CacheOnly ){
		lookCacheOnly(Conn,tc,fc);
		goto EXIT;
	}

	if( ImMaster && ACT_GENERALIST ){ /* metamo. */
		if( !do_RELAY(Conn,RELAY_DELEGATE)
		 && !do_RELAY(Conn,RELAY_PROXY) ){ 
			sv1log("Forbidden: RELAY by switch to GENERALIST\n");
			returnAckDENIED(Conn,tc,"relay forbidden");
			goto EXIT;
		}
	}

	returnAckOK(Conn,tc,"http");

	REQ[0] = 0;
	if( fgetsRequest(Conn,REQ,sizeof(REQ),fc,0) == NULL || REQ[0] == 0 ){
		if( RequestSerno ){
			clntClose(Conn,"q:by client (request EOF-8)");
			setClientEOF(Conn,fc,"request-EOF-8");
			goto EXIT;
		}
		gotNullRequest(Conn,QX,fc,tc);
		goto EXIT;
	}
	if( nonHTTP(Conn,ftc,REQ) )
		goto NHEXIT;

	if( HTTP11_clientTEST && ReqVer(REQ,'0',REQ,'1') ){
		sv1log("#HT11 emulate HTTP/1.1 client's request.\n");
	}

	HTTP_decompRequest(Conn);
	strcpy(OREQ,REQ);
	strcpy(OREQ_HOST,DFLT_HOST);
	OREQ_PORT = DFLT_PORT;

	if( ReqVer(REQ,'1',REQ,HTTP11_toserver?'1':'0') ){
		if( HTTP11_toclient )
		ClntAccChunk = 1;
		clientAskedKeepAlive(Conn,"Connection","keep-alive");
		if( HTTP11_toclient == 0 )
			WillKeepAlive = 0;
		/* This set-up for keep-alive must be done before interpreting
		 * Request-Header (Connection: field)
		 */
	}

	BadRequest = 0;
	Normalized = 0;
	recvRequestFields(Conn,QX,fc);
	if( 0 <= QX_range[0] ){
		sv1log("#HT11 Don't use chunked encoding for Range: %d-%d\n",
			QX_range[0],QX_range[1]);
		ClntAccChunk = 0;
	}

	if( BadRequest && HTTP_rejectBadHeader ){
		sv1log("##HHe ILLEGAL HEADER##\n");
		if( HTTP_methodWithBody(REQ_METHOD) ){
			FILE *xs = NULLFP();
			relayRequestBody1(Conn,fc,xs,xs,REQ,REQ_FIELDS);
		}
		lineScan("beforeConnnect",BadServDetected);
		lineScan("NOT-CONNECTED",BadServResponse);
		httpStat = CS_BADREQUEST;
		putBadRequest(Conn,tc);
		goto EXIT;
	}

	if( strncmp(REQ_URL,"data:",5) == 0 ){
		HTTP_putData(Conn,tc,REQ_VNO,REQ_URL+5);
		goto EXIT;
	}

	if( doCONNECT(Conn,QX,fc,tc) )
		goto EXIT;

	if( MountMoved(Conn,tc,REQ,REQ_URL) )
		goto EXIT;
	if( MountSpecialResponse(Conn,tc) )
		goto EXIT;
	if( doMaxHops(Conn,QX,fc,tc) )
		goto EXIT;

	HTTP_getHost(Conn,REQ,REQ_FIELDS);
	if( rewriteRequest(Conn,QX,fc,tc) < 0 )
		goto EXIT;

	if( DO_DELEGATE && delegate_moved(Conn,tc) )
		goto EXIT;

	setREQUEST(Conn,REQ);
	decompREQUEST(Conn,QX);

	if( RelayForbidden || HTTP_forgedAuthorization(Conn,REQ_FIELDS) ){
		QX_totalc = putHttpRejectmsg(Conn,tc,QX_proto,QX_site,QX_port,REQ);
		http_Log(Conn,403,CS_AUTHERR,REQ,QX_totalc);
		goto EXIT;
	}

	setupConnect(Conn);
	if( Conn->co_nonet ){
		DontReadCache = 0;
		DontWriteCache = 1;
	}

	if( doMyself(Conn,QX,fc,tc) )
		goto EXIT;

	if( checkAUTH(Conn,QX,fc,tc,QX_proto,QX_site,QX_port,0) != 0 )
		goto EXIT;

	if( findFieldValue(REQ_FIELDS,"Authorization") ){
		sv1log("Authorization: Dont-Read/Write-Cache ON\n");
		DontReadCache = 1;
		DontWriteCache = 1;
	}
	if( checkCookie(Conn,QX,fc,tc) < 0 )
		goto EXIT;

	tc = insertFTOCL_X(Conn,tc);
	setFTOCL(NULL);

	if( IAM_GATEWAY && ClntKeepAlive )
		if( !strcaseeq(DST_PROTO,"nntp")
		 && !strcaseeq(DST_PROTO,"news")
		 && !strcaseeq(DST_PROTO,"pop") )
			clntClose(Conn,"g:gateway for: %s",DST_PROTO);

	if( !DontReadCache && streq(QX_proto,"ftp") ){
		if( doFtpCached(Conn,QX,fc,tc) )
			goto EXIT;
	}

	if( GET_CACHE ){
		/*fclose(tc);*/
		tc = fopen("/dev/null","w");
	}

	if( QX_upath = strchr(REQ,'/') )
		QX_upath = QX_upath + 1;
	else	QX_upath = "?";

	if( IAM_GATEWAY
	 || 512 < (int)strlen(REQ)
	 || without_cache()
	 || strncasecmp(REQ,"GET ",4) != 0 ){
		QX_useCache = 0;
		QX_cpath[0] = 0;
		httpStat = CS_WITHOUTC;
	}else{
		QX_useCache =
		CTX_cache_path(Conn,QX_proto,QX_site,QX_port,QX_upath,QX_cpath);
		if( QX_useCache )
			httpStat = CS_NEW;
		else	httpStat = CS_WITHOUTC;
	}
	if( CacheOnly )
		QX_expire = 0x7FFFFFFF;
	else
	if( Conn->co_nonet )
		QX_expire = 0x7FFFFFFF;
	else	QX_expire = http_EXPIRE(Conn,QX_site);

	if( IAM_GATEWAY ){
	}else{
		if( doCache(Conn,QX,fc,tc) )
			goto EXIT;
		if( QX_cachefp == NULL )
			httpStat = CS_WITHOUTC;
		if( !reqRobotsTxt(Conn)  )
			genIfModified(Conn,QX_cachefp,QX_cpath,QX_cdate);
	}
	if( HTTP_reqWithHeader(Conn,REQ) ){
		HTTP_setHost(Conn,REQ_FIELDS);
	}
	if( HTTP_methodWithBody(REQ_METHOD) )
		pollRequestBody(Conn,QX,fc,tc);

	if( doHTTPMAIL(Conn,QX,fc,tc,QX_expire))
		goto EXIT;

	if( QX_cachefp != NULL )
		CacheMtime = file_mtime(fileno(QX_cachefp));

	setConnStart(Conn);
	FromS = -1;
	IsGopherSearch = 0;

	Conn->sv_reusing = 0;
	if( FromS < 0 ){
		if( Conn->sv_reusing = getServ(Conn) ) /* FromS is set ... */
			setConnDone(Conn);
	}

CONNECT_RETRY:

	if( FromS < 0 && tryConnects(Conn,QX,fc,tc) < 0 )
		goto EXIT;

	fs = fdopen(FromS,"r");
	if( ToServ )
		ts = ToServ;
	else{
		ts = fdopen(ToS,"w");
		if( ts == NULL ){
			sv1log("##FATAL: fdopen(ToS=%d) failed\n",ToS);
			goto EXIT;
		}
	}

	if( 0 <= Conn->ca_objsize && ConnType == 'i' ){
		/* response data is got by HIT_OBJ of ICP */
		QX_cpid = 0;
		QX_tsfd = -1;
		QX_fsfd = -1;
	}else{
		char *cpath = (QX_cachefp == NULL) ? NULL : QX_cpath;
		QX_tsfd = dup(fileno(ts));
		QX_fsfd = dup(fileno(fs));
		QX_cpid = relayRequest(Conn,QX,cpath,QX_cdate,fc,ts,fs);
		if( 0 < BadServ ){
			goto SVEXIT;
		}
		if( ferror(ts) || feof(fs) ){
			if( Conn->sv_reusing ){
			sv1log("#### SERVER REUSE: closed after request\n");
				closeServ(Conn,QX,ts,fs);
				goto CONNECT_RETRY;
			}
		}
	}

	if( DontWriteCache && QX_cachefp != NULL )
		CacheClose(QX);

	if( reqRobotsTxt(Conn) ){
		respFile = TMPFILE("HTTP-Reponse");
		tc_sav = tc;
		tc = respFile;
		chunksav = ClntAccChunk;
		ClntAccChunk = 0;
	}else	respFile = NULL;

	setRelaying(Conn,tc,QX_cachefp,QX_cpath);
	QX_rcode = relay_response(Conn,
			QX_cpid,
			QX_proto,QX_site,QX_port,
			REQ,
			QX_cpath,
			0,fs,tc,fc,
			QX_cachefp,
			QX_cdate != -1);

	if( respFile ){
		ClntAccChunk = chunksav;
		tc = tc_sav;
		fflush(respFile);
		fseek(respFile,0,0);
		if( QX_rcode != R_EMPTY_RESPONSE ){
			if( reqRobotsTxt(Conn) ){
				putRobotsTxt(Conn,tc_sav,respFile,1);
			}
		}
		fclose(respFile);
		respFile = 0;
	}

	if( QX_rcode == R_EMPTY_RESPONSE ){
		if( Conn->sv_reusing ){
			sv1log("#### SERVER REUSE: got empty response\n");
			closeServ(Conn,QX,ts,fs);
			goto CONNECT_RETRY;
		}
	}

	Verbose("relay_response()=%d, cache=%x, httpStat=%c DontTruncate=%d\n",
		QX_rcode,QX_cachefp,httpStat,DontTruncate);

	if( QX_cachefp != NULL )
		doTruncateCache(Conn,QX,fc,tc);

SVEXIT:
	if( 0 <= QX_tsfd ){
		int svkeep;
		if( 0 < BadServ )
			svkeep = 0;
		else
		if( httpStat == CS_EOF || 0 < fPollIn(fs,1) ){
			sv1log("#HT11 EOF or pending data from the server, httpStat=%c\n",httpStat);
			svkeep = 0;
		}else
		if( !feof(fs) && !ferror(ts) )
			svkeep = putServ(Conn,QX_tsfd,QX_fsfd);
		else	svkeep = 0;
		if( svkeep == 0 ){
			sv1log("#HT11 close svsokcs[%d,%d]\n",QX_tsfd,QX_fsfd);
			close(QX_tsfd);
			close(QX_fsfd);
			if( Conn->sv_reusing ){
				sv1log("#### SERVER REUSE: got EOF\n");
				delServ(Conn,fileno(ts),fileno(fs));
			}
		}
	}

	fclosesTIMEOUT(fs,ts);
	if( QX_cpid )
		wait(0);

EXIT:
	if( 0 < BadServ ){
		httpStat = CS_BADREQUEST;
		putBadRequest(Conn,tc);
	}

NHEXIT: /* NOTE:
	 * CurEnv might be released already after metamo. to Generalist
	 */

	ToServ = NULL;
	CacheClose(QX); /* unlock and remove AF_UNIX socket */
	if( QX_cpath[0] && File_size(QX_cpath) == 0 ){
		sv1log("unlink empty cache: %s\n",QX_cpath);
		QX_cacheRemove = 1;
	}
	if( QX_cacheRemove )
		CTX_cache_remove(Conn,QX_proto,QX_site,QX_port,QX_cpath);
	setRelaying(NULL);
	Where = 0;
}

/*
 * To cope with 305/306 response, current DeleGate implements it like this:
 * Whole request message including message body must be kept for
 * transparent automatic retry by the proxy, and two transactions
 * will occur always.
 * One another way is recording a triple (method,URL,proxy) for each
 * 305/306 response to be looked by succeeding proxy processes to decide
 * their routing.  But, it seems expensive too.
 */
static proxyRedirect(Conn,ftc)
	Connection *Conn;
	FILE *ftc[2];
{	int redirects;
	char proxy[1024];
	int comask;
	HtResp htResp;

	htResp = CurEnv->r_resp;
	for( redirects = 0; redirects < 4; redirects++ ){
		if( RespCode != 305 && RespCode != 306 )
			break;
		if( getFV(RESP_MSG,"Set-Proxy",proxy) == NULL )
			break;
		if( sizeof(OREQ_MSG) <= OREQ_LEN )
			break; /* incompletely buffered */
		if( strncasecmp(proxy,"DIRECT",5) == 0 ){
			sv1log("#### must be accessed without proxy: %s",OREQ);
			if( ConnType == 'p' || ConnType == 'm' ){
				ConnType = 0;
				DDI_pushCbuf(Conn,OREQ_MSG,OREQ_LEN);
				comask = Conn->co_mask;
				Conn->co_mask = (CONN_NOPROXY | CONN_NOMASTER);
				resetHTTPenv(Conn,CurEnv);
				service_http2(Conn,ftc);
				Conn->co_mask = comask;
				continue;
			}
		}
		break;
	}
	if( RESP_LEN == 0 ) /* redirection might be aborted */
		CurEnv->r_resp = htResp;

	if( RespCode == 305 || RespCode == 306 || ConnType == 0 )
		fputs(RESP_MSG,ftc[1]);

	/*
	 * if Proxy-Authorization is required, retry with it... ?
	 */
}

service_http(Conn)
	Connection *Conn;
{
	service_httpX(Conn);
}
service_httpX(Conn)
	Connection *Conn;
{	FILE *fc,*tc;
	FILE *ftc[2];
	void (*osigint)() = NULL;
	void (*osigterm)() = NULL;
	void (*osigpipe)() = NULL;
	char o_buff[OBUFSIZE];
	char i_buff[URLSZ];
	Connection savConn;
	int keepAlive,clntKeepAlive,Serno,tm0;
	int timeout,timeout1,timeout2,rtimeout,ptimeout,nready,till;
	HTTP_env httpEnv;
	char *FTOCL,*getFTOCL();
	FTOCL = getFTOCL(Conn);

	tc = fdopen(ToC,"w");
	if( tc == 0 ){
	http_Log(Conn,500,CS_ERROR,"ERR cannot_fdopen(to_client)",0);
	Exit(-1,"fdopen(ToC=%d) failed\n",ToC);
	}
	tcCLOSED = 0;

	fc = fdopen(FromC,"r");
	if( fc == NULL ){
	http_Log(Conn,500,CS_ERROR,"ERR cannot_fdopen(from_client)",0);
	Exit(-1,"fdopen(FromC=%d) failed\n",FromC);
	}

	ftc[0] = fc;
	ftc[1] = tc;

	/* setbuff tc after for IRIX-5.3 ... ?_? */
	setbuffer(fc,i_buff,sizeof(i_buff));
	setbuffer(tc,o_buff,sizeof(o_buff));

	osigint = Vsignal(SIGINT,sigTERM);
	osigterm = Vsignal(SIGTERM,sigTERM);
	osigpipe = Vsignal(SIGPIPE,sigPIPE);

	Conn->sv_dfltsav = Conn->sv_dflt;
	savConn = *Conn;
	keepAlive = 0;
	clntKeepAlive = 0;
	Serno = 0;
	WhyClosed[0] = 0;

	timeout = HTTP_CKA_TIMEOUT;
	CKA_RemAlive = HTTP_CKA_MAXREQ;

	if( 0 < CKA_RemAlive && Conn->cl_count > HTTP_CKA_PERCLIENT ){
		char clhost[256];
		if( getClientHostPort(Conn,clhost) == 0 )
			strcpy(clhost,"?");
		Verbose("HCKA: too many connections %d/%d for %s\n",
			Conn->cl_count,HTTP_CKA_PERCLIENT,clhost);
		sprintf(WhyClosed,"X:too many Keep-Alive (%s*%d)",clhost,Conn->cl_count);
		CKA_RemAlive = 0;
	}else
	if( 0 < CKA_RemAlive && Conn->cl_count == 1 && 1 < HTTP_CKA_PERCLIENT ){
		timeout = timeout * 10;
		if( timeout < 20 ) timeout = 20;
		if( 60 < timeout ) timeout = 60;
		CKA_RemAlive = CKA_RemAlive * 10;
		if( CKA_RemAlive < 20 ) CKA_RemAlive = 20;
		if( 60 < CKA_RemAlive ) CKA_RemAlive = 60;
	}
	rtimeout = timeout + HTTP_CKA_TIMEOUT_MARGIN;
	ptimeout = 20;
	if( rtimeout <= ptimeout ){
		timeout1 = rtimeout;
		timeout2 = 0;
	}else{
		timeout1 = ptimeout;
		timeout2 = rtimeout - ptimeout;
	}

	do {
		if( keepAlive ){
			/* second or later request in Keep-Alive mode */
			if( DDI_proceedFromC(Conn,fc) < 0 ){
				/* Flush possibly bufferd previous request in
				 * input buff. not read but peeked only.
				 * This must be done before clear_DGref() which
				 * will clear control data of the peeking.
				 */
				clntClose(Conn,"d:by client(request EOF-0)");
				setClientEOF(Conn,fc,"request-EOF-0");
				break;
			}
			clear_DGreq(Conn);
			restoreConn(Conn,&savConn);
		}
		setHTTPenv(Conn,&httpEnv);
		if( 0 < CKA_RemAlive ){
			setKeepAlive(Conn,timeout);
		}
		if( keepAlive ){
			if( feof(fc) ){
				clntClose(Conn,"d:by client(request EOF-1)");
				setClientEOF(Conn,fc,"request-EOF-1");
				break;
			}
			if( checkClientEOF(Conn,tc,"keep_alive") ){
				clntClose(Conn,"d:by client(request EOF-2)");
				break;
			}

			fflushKeepAlive(Conn,"request-EOF-3",fc,tc,0);
			if( ClientEOF ){
				clntClose(Conn,"d:by client(request EOF-3)");
				break;
			}
			/* if the flush of the response to the client is
			 * postponed here, it must be flushed when
			 * the connection to or the response from
			 * the next server is blocked... (may be in 
			 * service_http2())
			 */

			till = time(0) + timeout;

			tm0 = time(0);
			nready = DDI_PollIn(Conn,fc,timeout1*1000);
			if( nready < 0 ){
				clntClose(Conn,"d:by client(request EOF-4)");
				setClientEOF(Conn,fc,"request-EOF-4");
				break;
			}

			if( nready == 0 && 0 < timeout2 ){
				ProcTitle(Conn,"(HTTP:keep-alive=%02d:%02d)",
					(till%3600)/60,till%60);

				/* close ServPort not to block other clients
				 */
				checkCloseOnTimeout(0);
				LOG_flushall();
				nready = fPollIn(fc,timeout2*1000);
			}

			if( nready < 0 ){
				clntClose(Conn,"d:by client(request EOF-5)");
				setClientEOF(Conn,fc,"request-EOF-5");
				break;
			}
			if( nready == 0 ){
				clntClose(Conn,"t:timeout: %d",time(0)-tm0);
				break;
			}
			if( READYCC(fc) == 0 )
			if( Peek1(fileno(fc)) < 1 ){
				clntClose(Conn,"d:by client(request EOF-6)");
				setClientEOF(Conn,fc,"request-EOF-6");
				break;
			}

			/* close ServPort because this delegated may be
			 * devoted to a single client
			 */
			checkCloseOnTimeout(0);
		}

		strcpy(REQ_METHOD,"?");
		if( keepAlive )
			WhyClosed[0] = '-';
		else
		if( WhyClosed[0] == 0 )
			WhyClosed[0] = '?';
		SentKeepAlive = 0;
		ClntAccChunk = 0;
		setFTOCL(NULL);
		service_http2(Conn,ftc);
		if( ftc[0] == NULL || ftc[1] == NULL )
			break;

		if( RespCode == 305 || RespCode == 306 )
			proxyRedirect(Conn,ftc);

		CKA_RemAlive--;
		clntKeepAlive |= ClntKeepAlive;
		RES_CACHE_DISABLE = 0;

		if( tcCLOSED || ClientEOF )
			break;

		if( fflushTIMEOUT(tc) == EOF ){
			clntClose(Conn,"d:by client(request EOF-7)");
			setClientEOF(Conn,fc,"request-EOF-7");
			break;
		}

		if( 0 < CKA_RemAlive ){
			if( !streq(REQ_METHOD,"?") ) /* not empty request */
				Serno = incRequestSerno(Conn);
			if( keepAlive || WillKeepAlive )
				Verbose("HCKA:[%d] KeepAlive: %s %c =>%d\n",
					Serno,REQ_METHOD,httpStat,WillKeepAlive);
		}
		keepAlive = WillKeepAlive && SentKeepAlive;
	} while( keepAlive && 0 < CKA_RemAlive );
	setHTTPenv(Conn,NULL);

	if( clntKeepAlive && 1 < HTTP_CKA_MAXREQ ){
		http_logplus(Conn,WhyClosed[0]);
		sv1log("HCKA:[%d] closed -- %s\n",RequestSerno,WhyClosed);
	}

	Where = "closeClient";

	if( ftc[0] == NULL || ftc[1] == NULL ){
		sv1log("fclose() already: fc=%X/%d tc=%X/%d\n",
			fc,ftc[0],tc,ftc[1]);
	}else
	if( tcCLOSED )
		fclose(fc);
	else{
		if( checkClientEOF(Conn,tc,NULL) ){
			int fd = fileno(tc);
			close(fd);
			Verbose("## [%d] don't flush to cause SIGPIPE\n",fd);
		}else{
			if( 0 < DELEGATE_LINGER )
				set_linger(fileno(tc),DELEGATE_LINGER);
		}
		fclosesTIMEOUT(tc,fc); /* close tc first */
	}
	Vsignal(SIGPIPE,osigpipe);
	Vsignal(SIGTERM,osigterm);
	Vsignal(SIGINT,osigint);

	setFTOCL(FTOCL);
}
service_https(Conn)
	Connection *Conn;
{
	if( 0 <= ToS )
		relay_svcl(Conn,FromC,ToC,FromS,ToS);
	else	service_http(Conn);
}

static sendHttpResponse(Conn,in,out,req)
	Connection *Conn;
	FILE *in,*out;
	char *req;
{	char myhost[128];
	int myport;

	if( DONT_REWRITE ){
		strcpy(myhost,MYSELF);
		myport = 0;
	}else{	
		myport = ClientIF_H(Conn,myhost);
	}
	ImResponseFilter = 1;
	relay_response(Conn,0,CLNT_PROTO,myhost,myport,req,NULL,0,
		in,out,NULL,NULL,0);
}

FILE *openHttpResponseFilter(Conn,tc)
	Connection *Conn;
	FILE *tc;
{
	sv1log("## openHttpResponseFilter: clnt=%d will=%d chunk=%d\n",
		ClntKeepAlive,WillKeepAlive,ClntAccChunk);
	if( !ClntAccChunk ){
		/* URLs (in /-/ page for example) may be rewriten in the filter
		 * without correcting Content-Length
		 */
		HTTP_clntClose(Conn,"x:response filter");
	}
	Conn->fi_builtin = 1;
	return openFilter(Conn,"HttpResponseFilter",sendHttpResponse,tc,OREQ);
}
relayThru(Conn)
	Connection *Conn;
{
	return DONT_REWRITE && !cur_codeconvCL(NULL) && !CCXactive(CCX_TOCL);
}
service_file(Conn)
	Connection *Conn;
{
	sv1log("access to file ... \n");
}

FILE * CTX_URLget(OrigConn,origctx,url,reload,out)
	Connection *OrigConn;
	char *url;
	FILE *out;
{	Connection ConnBuf,*Conn = &ConnBuf;
	char req[1024],*rp,resp[1024];
	int io[2];
	FILE *aout;
	HttpResponse resx;

	aout = out;
	if( out == NULL )
		out = TMPFILE("CTX_URLget");

	if( strncasecmp(url,"file:",5) == 0 )
	if( !isExecutableURL(url) ){
		FILE *fp;
		char host[128],*path;

		path = file_hostpath(url,NULL,host);
		if( fp = fopen(path,"r") ){
			copyfile1(fp,out);
			fclose(fp);
			fflush(out);
			Ftruncate(out,0,1);
			fseek(out,0,0);
			return out;
		}
		if( aout == NULL && out != NULL )
			fclose(out);
		return NULL;
	}
	if( strncasecmp(url,"data:",5) == 0 ){
		HTTP_putData(OrigConn,out,0,url+5);
		fflush(out);
		Ftruncate(out,0,1);
		fseek(out,0,0);
		return out;
	}
	if( strncasecmp(url,"builtin:",8) == 0 ){
		extern char *getMssg();
		char *data;
		int leng;

		sprintf(req,"builtin/%s",url+8);
		data = getMssg(req,&leng);
		if( data == NULL )
			return NULL;
/*
 * eval. if with ".dhtml" extension.
 * on each request based on conditional info.
 */
		fwrite(data,1,leng,out);
		fflush(out);
		Ftruncate(out,0,1);
		fseek(out,0,0);
		return out;
	}

	if( OrigConn && origctx ){
		ConnCopy(Conn,OrigConn);
		ClntKeepAlive = 0;
		WillKeepAlive = 0;
		ToMyself = 0;
		Conn->sv_dflt = OrigConn->sv_dfltsav;
		bzero(&Conn->sv,sizeof(Conn->sv));
	}else{
		ConnInit(Conn);
		Conn->from_myself = 1;
	}
	ACT_SPECIALIST = 1;

	if( OrigConn ){
		ClientSock = OrigConn->cl_sockFd;
		/*
		Conn->oc_isset = 1;
		Conn->oc_norewrite = OrigConn->oc_norewrite;
		strcpy(Conn->oc_proxy,OrigConn->oc_proxy);
		strcpy(Conn->oc_proto,OrigConn->oc_proto);
		*/
	}

	Socketpair(io);
	FromC = io[0];
	ToC = dup(fileno(out));

	rp = Sprintf(req,"GET %s HTTP/1.0\r\n",url);
	rp = Sprintf(rp,"User-Agent: DeleGate/%s\r\n",DELEGATE_ver());
	if( reload )
		rp = Sprintf(rp,"Pragma: no-cache\r\n");
	strcpy(rp,"\r\n");

	DDI_pushCbuf(Conn,req,strlen(req));
	service_http(Conn);

	fflush(out);
	Ftruncate(out,0,1);
	fseek(out,0,0);

	if( fgets(resp,sizeof(resp),out) != NULL ){
		if( strncmp(resp,F_HTTPVER,F_HTTPVERLEN) == 0 ){
			if( decomp_http_status(resp,&resx) != 200 ){
				Ftruncate(out,0,0);
				fseek(out,0,0);
				sv1log("error: %s",resp);
			}else{
				while( fgets(resp,sizeof(resp),out) != NULL )
					if( resp[0] == '\r' || resp[0] == '\n' )
						break;
			}
		}else{
			fseek(out,0,0);
		}
	}

	close(io[0]);
	close(io[1]);

	return out;
}
FILE * URLget(url,reload,out)
	char *url;
	FILE *out;
{
	return CTX_URLget(NULL,0,url,reload,out);
}

extern char *TIMEFORM_RFC822;
putFileInHTTP(tc,path,file)
	FILE *tc;
	char *path,*file;
{	FILE *fp;
	int size,mtime;
	char sdate[1024];

	fp = fopen(path,"r");
	if( fp == NULL ){
		fprintf(tc,"HTTP/1.0 404 not found\r\n");
		return;
	}
	size = file_size(fileno(fp));
	mtime = file_mtime(fileno(fp));
	StrftimeGMT(sdate,sizeof(sdate),TIMEFORM_RFC822,mtime);

	fprintf(tc,"HTTP/1.0 200 OK\r\n");
	fprintf(tc,"MIME-Version: 1.0\r\n");
	fprintf(tc,"Content-Type: application/octet-stream\r\n");
	fprintf(tc,"Content-Transfer-Encoding: base64\r\n");
	fprintf(tc,"Content-Length: %d\r\n",size);
	fprintf(tc,"Last-Modified: %s\r\n",sdate);
	fprintf(tc,"\r\n");
	MIME_to64(fp,tc);
	fclose(fp);
}
